# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains the manager for handler methods used by the callback client"""
import logging
import threading
import abc
from azure.iot.device.common import handle_exceptions
from azure.iot.device.iothub.client_event import (
    CONNECTION_STATE_CHANGE,
    NEW_SASTOKEN_REQUIRED,
    BACKGROUND_EXCEPTION,
)
import concurrent.futures

logger = logging.getLogger(__name__)

# Receiver Handlers
MESSAGE = "_on_message_received"
METHOD = "_on_method_request_received"
TWIN_DP_PATCH = "_on_twin_desired_properties_patch_received"

# Client Event Handler Runner
CLIENT_EVENT = "client_event"
# Client Event Names
client_events = [CONNECTION_STATE_CHANGE, NEW_SASTOKEN_REQUIRED, BACKGROUND_EXCEPTION]


class HandlerManagerException(Exception):
    """An exception raised by a HandlerManager"""

    pass


class HandlerRunnerKillerSentinel(object):
    """An object that functions according to the sentinel design pattern.
    Insert into an Inbox in order to indicate that the Handler Runner associated with that
    Inbox should be stopped.
    """

    pass


class AbstractHandlerManager(abc.ABC):
    """Partial class that defines handler manager functionality shared between sync/async"""

    def __init__(self, inbox_manager):
        self._inbox_manager = inbox_manager

        self._receiver_handler_runners = {
            MESSAGE: None,
            METHOD: None,
            TWIN_DP_PATCH: None,
        }
        self._client_event_runner = None

        # Receiver handlers (Each will have it's own runner)
        self._on_message_received = None
        self._on_method_request_received = None
        self._on_twin_desired_properties_patch_received = None

        # Client Event handlers (Share a single Client Event runner)
        self._on_connection_state_change = None
        self._on_new_sastoken_required = None
        self._on_background_exception = None

        # As mentioned above, Receiver handlers each get their own runner. This is because it is
        # reasonably possible that many receives (and many different types of receives) can be
        # happening in a very short time. This is because we want to be able to support processing
        # multiple receives simultaneously, but also protect different receive invocations from
        # being slowed by a poorly written handler of a different type. Thus each Receiver handler
        # gets its own unique runner that cannot end up blocked by other handlers.
        #
        # Client Events on the other hand are generated by the client itself rather than
        # unsolicited data received over the wire. As these are relatively uncommon, we don't
        # expect multiple Client Events of the same type to really start queueing up, so there
        # isn't much justification for each type of event getting it's own dedicated runner.
        # Furthermore, there is much less concern over one inefficient or slow handler blocking
        # execution of others due to this infrequency.
        #
        # However, there are other differences between these handler classes as well. Receiver
        # handlers will always be invoked with a single argument - the received data structure.
        # Client Event handlers on the other hand are more flexible - they may be invoked with
        # different numbers of arguments depending on the handler, from none to multiple. This
        # is to keep design space open for more complex Client Events in the future.
        #
        # Finally, it is possible to stop ONLY the receiver handlers via the .stop() method
        # if desired. This is useful because while receiver handlers should be stopped when the
        # client disconnects, Client Events can still occur while the client is disconnected
        # as they are propagated from the client itself rather than received data.

    def _get_inbox_for_receive_handler(self, handler_name):
        """Retrieve the inbox relevant to the handler"""
        if handler_name == METHOD:
            return self._inbox_manager.get_method_request_inbox()
        elif handler_name == TWIN_DP_PATCH:
            return self._inbox_manager.get_twin_patch_inbox()
        elif handler_name == MESSAGE:
            return self._inbox_manager.get_unified_message_inbox()
        else:
            return None

    def _get_handler_for_client_event(self, event_name):
        """Retrieve the handler relevant to the event"""
        if event_name == NEW_SASTOKEN_REQUIRED:
            return self._on_new_sastoken_required
        elif event_name == CONNECTION_STATE_CHANGE:
            return self._on_connection_state_change
        elif event_name == BACKGROUND_EXCEPTION:
            return self._on_background_exception
        else:
            return None

    @abc.abstractmethod
    def _receiver_handler_runner(self, inbox, handler_name):
        """Run infinite loop that waits for an inbox to receive an object from it, then calls
        the handler with that object
        """
        pass

    @abc.abstractmethod
    def _client_event_handler_runner(self, handler_name):
        """Run infinite loop that waits for the client event inbox to receive an event from it,
        then calls the handler that corresponds to that event
        """
        pass

    @abc.abstractmethod
    def _start_handler_runner(self, handler_name):
        """Create, and store a handler runner"""
        pass

    @abc.abstractmethod
    def _stop_receiver_handler_runner(self, handler_name):
        """Cancel and remove a handler runner"""
        pass

    @abc.abstractmethod
    def _stop_client_event_handler_runner(self):
        """Cancel the client event handler runner"""
        pass

    def _generic_receiver_handler_setter(self, handler_name, new_handler):
        """Set a handler"""
        curr_handler = getattr(self, handler_name)
        if new_handler is not None and curr_handler is None:
            # Create runner, set handler
            logger.debug("Creating new handler runner for handler: {}".format(handler_name))
            setattr(self, handler_name, new_handler)
            self._start_handler_runner(handler_name)
        elif new_handler is None and curr_handler is not None:
            # Cancel runner, remove handler
            logger.debug("Removing handler runner for handler: {}".format(handler_name))
            self._stop_receiver_handler_runner(handler_name)
            setattr(self, handler_name, new_handler)
        else:
            # Update handler, no need to change runner
            logger.debug("Updating set handler: {}".format(handler_name))
            setattr(self, handler_name, new_handler)

    @staticmethod
    def _generate_callback_for_handler(handler_name):
        """Define a callback that can handle errors during handler execution"""

        def handler_callback(future):
            try:
                e = future.exception(timeout=0)
            except Exception as raised_e:
                # This shouldn't happen because cancellation or timeout shouldn't occur...
                # But just in case...
                new_err = HandlerManagerException(
                    "HANDLER ({}): Unable to retrieve exception data from incomplete invocation".format(
                        handler_name
                    )
                )
                new_err.__cause__ = raised_e
                handle_exceptions.handle_background_exception(new_err)
            else:
                if e:
                    new_err = HandlerManagerException(
                        "HANDLER ({}): Error during invocation".format(handler_name),
                    )
                    new_err.__cause__ = e
                    handle_exceptions.handle_background_exception(new_err)
                else:
                    logger.debug(
                        "HANDLER ({}): Successfully completed invocation".format(handler_name)
                    )

        return handler_callback

    def stop(self, receiver_handlers_only=False):
        """Stop the process of invoking handlers in response to events.
        All pending items will be handled prior to stoppage.
        """
        # Stop receiver handlers
        for handler_name in self._receiver_handler_runners:
            if self._receiver_handler_runners[handler_name] is not None:
                self._stop_receiver_handler_runner(handler_name)

        # Stop the client event handler (if instructed)
        if not receiver_handlers_only and self._client_event_runner is not None:
            self._stop_client_event_handler_runner()

    def ensure_running(self):
        """Ensure the process of invoking handlers in response to events is running"""
        # Ensure any receiver handler set on the manager has a corresponding handler runner running
        for handler_name in self._receiver_handler_runners:
            if (
                self._receiver_handler_runners[handler_name] is None
                and getattr(self, handler_name) is not None
            ):
                self._start_handler_runner(handler_name)

        # Ensure client event handler runner is running if at least one client event handler is set
        # on the manager
        if self._client_event_runner is None:
            for event in client_events:
                handler = self._get_handler_for_client_event(event)
                if handler is not None:
                    self._start_handler_runner(CLIENT_EVENT)
                    break

    # ~~~Receiver Handlers~~~
    # Setting a receiver handler will start a dedicated runner for that handler
    # Removing a receiver handler will stop the dedicated runner for that handler
    @property
    def on_message_received(self):
        return self._on_message_received

    @on_message_received.setter
    def on_message_received(self, value):
        self._generic_receiver_handler_setter(MESSAGE, value)

    @property
    def on_method_request_received(self):
        return self._on_method_request_received

    @on_method_request_received.setter
    def on_method_request_received(self, value):
        self._generic_receiver_handler_setter(METHOD, value)

    @property
    def on_twin_desired_properties_patch_received(self):
        return self._on_twin_desired_properties_patch_received

    @on_twin_desired_properties_patch_received.setter
    def on_twin_desired_properties_patch_received(self, value):
        self._generic_receiver_handler_setter(TWIN_DP_PATCH, value)

    # ~~~Client Event Handlers~~~
    # Setting any client event handler will start the shared client event handler runner
    # Removing handlers will NOT stop the client event handler runner - you must use .stop()
    # Stopping when all client event handlers are removed could be added if necessary.
    @property
    def on_connection_state_change(self):
        return self._on_connection_state_change

    @on_connection_state_change.setter
    def on_connection_state_change(self, value):
        self._on_connection_state_change = value
        if self._client_event_runner is None:
            self._start_handler_runner(CLIENT_EVENT)

    @property
    def on_new_sastoken_required(self):
        return self._on_new_sastoken_required

    @on_new_sastoken_required.setter
    def on_new_sastoken_required(self, value):
        self._on_new_sastoken_required = value
        if self._client_event_runner is None:
            self._start_handler_runner(CLIENT_EVENT)

    @property
    def on_background_exception(self):
        return self._on_background_exception

    @on_background_exception.setter
    def on_background_exception(self, value):
        self._on_background_exception = value
        if self._client_event_runner is None:
            self._start_handler_runner(CLIENT_EVENT)

    # ~~~Other Properties~~~
    @property
    def handling_client_events(self):
        """Indicates if the HandlerManager is currently capable of resolving ClientEvents"""
        # This client event runner is only running if at least one handler for client events has
        # been set. If none have been set, it is dangerous to add items to the client event inbox
        # as none will ever be retrieved due to no runner process occurring, thus the need for this
        # check.
        #
        # The ideal solution would be to always keep the client event runner running, but this
        # could break older customer code due to older APIs on the customer-facing clients. It is
        # unfortunate that something related to an API has seeped into this internal and ideally
        # isolated module, but the needs of the client design have influenced the design of this
        # manager (by only starting the runner when a handler is set), so the mitigation must also
        # be located in this module.
        if self._client_event_runner is None:
            return False
        else:
            return True


class SyncHandlerManager(AbstractHandlerManager):
    """Handler manager for use with synchronous clients"""

    def _receiver_handler_runner(self, inbox, handler_name):
        """Run infinite loop that waits for an inbox to receive an object from it, then calls
        the handler with that object
        """
        logger.debug("HANDLER RUNNER ({}): Starting runner".format(handler_name))
        _handler_callback = self._generate_callback_for_handler(handler_name)

        # Run the handler in a threadpool, so that it cannot block other handlers (from a different task),
        # or the main client thread. The number of worker threads forms an upper bound on how many instances
        # of the same handler can be running simultaneously.
        tpe = concurrent.futures.ThreadPoolExecutor(max_workers=4)
        while True:
            handler_arg = inbox.get()
            if isinstance(handler_arg, HandlerRunnerKillerSentinel):
                # Exit the runner when a HandlerRunnerKillerSentinel is found
                logger.debug(
                    "HANDLER RUNNER ({}): HandlerRunnerKillerSentinel found in inbox. Exiting.".format(
                        handler_name
                    )
                )
                tpe.shutdown()
                break
            # NOTE: we MUST use getattr here using the handler name, as opposed to directly passing
            # the handler in order for the handler to be able to be updated without cancelling
            # the running task created for this coroutine
            handler = getattr(self, handler_name)
            logger.debug("HANDLER RUNNER ({}): Invoking handler".format(handler_name))
            fut = tpe.submit(handler, handler_arg)
            fut.add_done_callback(_handler_callback)
            # Free up this object so the garbage collector can free it if necessary. If we don't
            # do this, we end up keeping this object alive until the next event arrives, which
            # might be a long time. Tests would flag this as a memory leak if that happened.
            del handler_arg

    def _client_event_handler_runner(self):
        """Run infinite loop that waits for the client event inbox to receive an event from it,
        then calls the handler that corresponds to that event
        """
        logger.debug("HANDLER RUNNER (CLIENT EVENT): Starting runner")
        _handler_callback = self._generate_callback_for_handler("CLIENT_EVENT")

        tpe = concurrent.futures.ThreadPoolExecutor(max_workers=4)
        event_inbox = self._inbox_manager.get_client_event_inbox()
        while True:
            event = event_inbox.get()
            if isinstance(event, HandlerRunnerKillerSentinel):
                # Exit the runner when a HandlerRunnerKillerSentinel is found
                logger.debug(
                    "HANDLER RUNNER (CLIENT EVENT): HandlerRunnerKillerSentinel found in event queue. Exiting."
                )
                tpe.shutdown()
                break
            handler = self._get_handler_for_client_event(event.name)
            if handler is not None:
                logger.debug(
                    "HANDLER RUNNER (CLIENT EVENT): {} event received. Invoking {} handler".format(
                        event, handler
                    )
                )
                fut = tpe.submit(handler, *event.args_for_user)
                fut.add_done_callback(_handler_callback)
                # Free up this object so the garbage collector can free it if necessary. If we don't
                # do this, we end up keeping this object alive until the next event arrives, which
                # might be a long time. Tests would flag this as a memory leak if that happened.
                del event
            else:
                logger.debug(
                    "No handler for event {} set. Skipping handler invocation".format(event)
                )

    def _start_handler_runner(self, handler_name):
        """Start and store a handler runner thread"""
        # Client Event handler flow
        if handler_name == CLIENT_EVENT:
            if self._client_event_runner is not None:
                # This branch of code should NOT be reachable due to checks prior to the invocation
                # of this method. The branch exists for safety.
                raise HandlerManagerException(
                    "Cannot create thread for handler runner: {}. Runner thread already exists".format(
                        handler_name
                    )
                )
            # Client events share a handler
            thread = threading.Thread(target=self._client_event_handler_runner)
            # Store the thread
            self._client_event_runner = thread

        # Receiver handler flow
        else:
            if self._receiver_handler_runners[handler_name] is not None:
                # This branch of code should NOT be reachable due to checks prior to the invocation
                # of this method. The branch exists for safety.
                raise HandlerManagerException(
                    "Cannot create thread for handler runner: {}. Runner thread already exists".format(
                        handler_name
                    )
                )
            inbox = self._get_inbox_for_receive_handler(handler_name)
            # Each receiver handler gets its own runner
            thread = threading.Thread(
                target=self._receiver_handler_runner, args=[inbox, handler_name]
            )
            # Store the thread
            self._receiver_handler_runners[handler_name] = thread

        # NOTE: It would be nice to have some kind of mechanism for making sure this thread
        # doesn't crash or raise errors, but it would require significant extra infrastructure
        # and an exception in here isn't supposed to happen anyway. Perhaps it could be added
        # later if truly necessary
        thread.daemon = True  # Don't block program exit
        thread.start()

    def _stop_receiver_handler_runner(self, handler_name):
        """Stop and remove a handler runner thread.
        All pending items in the corresponding inbox will be handled by the handler before stoppage.
        """
        logger.debug(
            "Adding HandlerRunnerKillerSentinel to inbox corresponding to {} handler runner".format(
                handler_name
            )
        )
        inbox = self._get_inbox_for_receive_handler(handler_name)
        inbox.put(HandlerRunnerKillerSentinel())

        # Wait for Handler Runner to end due to the sentinel
        logger.debug("Waiting for {} handler runner to exit...".format(handler_name))
        thread = self._receiver_handler_runners[handler_name]
        thread.join()
        self._receiver_handler_runners[handler_name] = None
        logger.debug("Handler runner for {} has been stopped".format(handler_name))

    def _stop_client_event_handler_runner(self):
        """Stop and remove a handler runner thread.
        All pending items in the client event queue will be handled by handlers (if they exist)
        before stoppage.
        """
        logger.debug("Adding HandlerRunnerKillerSentinel to client event queue")
        event_inbox = self._inbox_manager.get_client_event_inbox()
        event_inbox.put(HandlerRunnerKillerSentinel())

        # Wait for Handler Runner to end due to the stop command
        logger.debug("Waiting for client event handler runner to exit...")
        thread = self._client_event_runner
        thread.join()
        self._client_event_runner = None
        logger.debug("Handler runner for client events has been stopped")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
This module contains user-facing asynchronous Provisioning Device Client for Azure Provisioning
Device SDK. This client uses Symmetric Key and X509 authentication to register devices with an
IoT Hub via the Device Provisioning Service.
"""
from __future__ import annotations  # Needed for annotation bug < 3.10
import logging
from typing import Any
from azure.iot.device.common import async_adapter
from azure.iot.device.custom_typing import FunctionOrCoroutine
from azure.iot.device.provisioning.abstract_provisioning_device_client import (
    AbstractProvisioningDeviceClient,
)
from azure.iot.device.provisioning.abstract_provisioning_device_client import (
    log_on_register_complete,
)
from azure.iot.device.provisioning.pipeline import exceptions as pipeline_exceptions
from azure.iot.device import exceptions
from azure.iot.device.provisioning.pipeline import constant as dps_constant
from azure.iot.device.provisioning.models import RegistrationResult

logger = logging.getLogger(__name__)


async def handle_result(callback: FunctionOrCoroutine[[Any], Any]) -> Any:
    try:
        return await callback.completion()
    except pipeline_exceptions.ConnectionDroppedError as e:
        raise exceptions.ConnectionDroppedError("Lost connection to IoTHub") from e
    except pipeline_exceptions.ConnectionFailedError as e:
        raise exceptions.ConnectionFailedError("Could not connect to IoTHub") from e
    except pipeline_exceptions.UnauthorizedError as e:
        raise exceptions.CredentialError("Credentials invalid, could not connect") from e
    except pipeline_exceptions.ProtocolClientError as e:
        raise exceptions.ClientError("Error in the IoTHub client") from e
    except pipeline_exceptions.OperationTimeout as e:
        raise exceptions.OperationTimeout("Could not complete operation before timeout") from e
    except pipeline_exceptions.PipelineNotRunning as e:
        raise exceptions.ClientError("Client has already been shut down") from e
    except Exception as e:
        raise exceptions.ClientError("Unexpected failure") from e


class ProvisioningDeviceClient(AbstractProvisioningDeviceClient):
    """
    Client which can be used to run the registration of a device with provisioning service
    using Symmetric Key or X509 authentication.
    """

    async def register(self) -> RegistrationResult:
        """
        Register the device with the provisioning service.

        Before returning the client will also disconnect from the provisioning service.
        If a registration attempt is made while a previous registration is in progress it may
        throw an error.

        Once the device is successfully registered, the client will no longer be operable.

        :returns: RegistrationResult indicating the result of the registration.
        :rtype: :class:`azure.iot.device.RegistrationResult`

        :raises: :class:`azure.iot.device.exceptions.CredentialError` if credentials are invalid
            and a connection cannot be established.
        :raises: :class:`azure.iot.device.exceptions.ConnectionFailedError` if a establishing a
            connection results in failure.
        :raises: :class:`azure.iot.device.exceptions.ConnectionDroppedError` if connection is lost
            during execution.
        :raises: :class:`azure.iot.device.exceptions.ClientError` if there is an unexpected failure
            during execution.

        """
        logger.info("Registering with Provisioning Service...")

        if not self._pipeline.responses_enabled[dps_constant.REGISTER]:
            await self._enable_responses()

        register_async = async_adapter.emulate_async(self._pipeline.register)
        register_complete = async_adapter.AwaitableCallback(return_arg_name="result")
        await register_async(payload=self._provisioning_payload, callback=register_complete)
        result = await handle_result(register_complete)

        log_on_register_complete(result)

        if result is not None and result.status == "assigned":
            logger.debug("Beginning pipeline shutdown operation")
            shutdown_async = async_adapter.emulate_async(self._pipeline.shutdown)
            callback = async_adapter.AwaitableCallback()
            await shutdown_async(callback=callback)
            await handle_result(callback)
            logger.debug("Completed pipeline shutdown operation")

        return result

    async def _enable_responses(self) -> None:
        """Enable to receive responses from Device Provisioning Service."""
        logger.info("Enabling reception of response from Device Provisioning Service...")
        subscribe_async = async_adapter.emulate_async(self._pipeline.enable_responses)

        subscription_complete = async_adapter.AwaitableCallback()
        await subscribe_async(callback=subscription_complete)
        await handle_result(subscription_complete)

        logger.info("Successfully subscribed to Device Provisioning Service to receive responses")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module provides hard coded patches used to modify items from the libraries.
Currently we have to do like this so that we don't use exec anywhere"""


def execute_patch_for_sync():
    from azure.iot.device.iothub.sync_clients import IoTHubDeviceClient as IoTHubDeviceClient

    def connect(self):
        return super(IoTHubDeviceClient, self).connect()

    connect.__doc__ = IoTHubDeviceClient.connect.__doc__
    setattr(IoTHubDeviceClient, "connect", connect)

    def disconnect(self):
        return super(IoTHubDeviceClient, self).disconnect()

    disconnect.__doc__ = IoTHubDeviceClient.disconnect.__doc__
    setattr(IoTHubDeviceClient, "disconnect", disconnect)

    def get_twin(self):
        return super(IoTHubDeviceClient, self).get_twin()

    get_twin.__doc__ = IoTHubDeviceClient.get_twin.__doc__
    setattr(IoTHubDeviceClient, "get_twin", get_twin)

    def patch_twin_reported_properties(self, reported_properties_patch):
        return super(IoTHubDeviceClient, self).patch_twin_reported_properties(
            reported_properties_patch
        )

    patch_twin_reported_properties.__doc__ = (
        IoTHubDeviceClient.patch_twin_reported_properties.__doc__
    )
    setattr(IoTHubDeviceClient, "patch_twin_reported_properties", patch_twin_reported_properties)

    def receive_method_request(self, method_name=None, block=True, timeout=None):
        return super(IoTHubDeviceClient, self).receive_method_request(method_name, block, timeout)

    receive_method_request.__doc__ = IoTHubDeviceClient.receive_method_request.__doc__
    setattr(IoTHubDeviceClient, "receive_method_request", receive_method_request)

    def receive_twin_desired_properties_patch(self, block=True, timeout=None):
        return super(IoTHubDeviceClient, self).receive_twin_desired_properties_patch(block, timeout)

    receive_twin_desired_properties_patch.__doc__ = (
        IoTHubDeviceClient.receive_twin_desired_properties_patch.__doc__
    )
    setattr(
        IoTHubDeviceClient,
        "receive_twin_desired_properties_patch",
        receive_twin_desired_properties_patch,
    )

    def send_message(self, message):
        return super(IoTHubDeviceClient, self).send_message(message)

    send_message.__doc__ = IoTHubDeviceClient.send_message.__doc__
    setattr(IoTHubDeviceClient, "send_message", send_message)

    def send_method_response(self, method_response):
        return super(IoTHubDeviceClient, self).send_method_response(method_response)

    send_method_response.__doc__ = IoTHubDeviceClient.send_method_response.__doc__
    setattr(IoTHubDeviceClient, "send_method_response", send_method_response)

    def shutdown(self):
        return super(IoTHubDeviceClient, self).shutdown()

    shutdown.__doc__ = IoTHubDeviceClient.shutdown.__doc__
    setattr(IoTHubDeviceClient, "shutdown", shutdown)

    def update_sastoken(self, sastoken):
        return super(IoTHubDeviceClient, self).update_sastoken(sastoken)

    update_sastoken.__doc__ = IoTHubDeviceClient.update_sastoken.__doc__
    setattr(IoTHubDeviceClient, "update_sastoken", update_sastoken)

    def create_from_connection_string(cls, connection_string, **kwargs):
        return super(IoTHubDeviceClient, cls).create_from_connection_string(
            connection_string, **kwargs
        )

    create_from_connection_string.__doc__ = IoTHubDeviceClient.create_from_connection_string.__doc__
    setattr(
        IoTHubDeviceClient,
        "create_from_connection_string",
        classmethod(create_from_connection_string),
    )

    def create_from_sastoken(cls, sastoken, **kwargs):
        return super(IoTHubDeviceClient, cls).create_from_sastoken(sastoken, **kwargs)

    create_from_sastoken.__doc__ = IoTHubDeviceClient.create_from_sastoken.__doc__
    setattr(IoTHubDeviceClient, "create_from_sastoken", classmethod(create_from_sastoken))

    def create_from_symmetric_key(cls, symmetric_key, hostname, device_id, **kwargs):
        return super(IoTHubDeviceClient, cls).create_from_symmetric_key(
            symmetric_key, hostname, device_id, **kwargs
        )

    create_from_symmetric_key.__doc__ = IoTHubDeviceClient.create_from_symmetric_key.__doc__
    setattr(IoTHubDeviceClient, "create_from_symmetric_key", classmethod(create_from_symmetric_key))

    def create_from_x509_certificate(cls, x509, hostname, device_id, **kwargs):
        return super(IoTHubDeviceClient, cls).create_from_x509_certificate(
            x509, hostname, device_id, **kwargs
        )

    create_from_x509_certificate.__doc__ = IoTHubDeviceClient.create_from_x509_certificate.__doc__
    setattr(
        IoTHubDeviceClient,
        "create_from_x509_certificate",
        classmethod(create_from_x509_certificate),
    )
    setattr(IoTHubDeviceClient, "connected", IoTHubDeviceClient.connected)
    setattr(
        IoTHubDeviceClient, "on_background_exception", IoTHubDeviceClient.on_background_exception
    )
    setattr(
        IoTHubDeviceClient,
        "on_connection_state_change",
        IoTHubDeviceClient.on_connection_state_change,
    )
    setattr(IoTHubDeviceClient, "on_message_received", IoTHubDeviceClient.on_message_received)
    setattr(
        IoTHubDeviceClient,
        "on_method_request_received",
        IoTHubDeviceClient.on_method_request_received,
    )
    setattr(
        IoTHubDeviceClient, "on_new_sastoken_required", IoTHubDeviceClient.on_new_sastoken_required
    )
    setattr(
        IoTHubDeviceClient,
        "on_twin_desired_properties_patch_received",
        IoTHubDeviceClient.on_twin_desired_properties_patch_received,
    )
    from azure.iot.device.iothub.sync_clients import IoTHubModuleClient as IoTHubModuleClient

    def connect(self):
        return super(IoTHubModuleClient, self).connect()

    connect.__doc__ = IoTHubModuleClient.connect.__doc__
    setattr(IoTHubModuleClient, "connect", connect)

    def disconnect(self):
        return super(IoTHubModuleClient, self).disconnect()

    disconnect.__doc__ = IoTHubModuleClient.disconnect.__doc__
    setattr(IoTHubModuleClient, "disconnect", disconnect)

    def get_twin(self):
        return super(IoTHubModuleClient, self).get_twin()

    get_twin.__doc__ = IoTHubModuleClient.get_twin.__doc__
    setattr(IoTHubModuleClient, "get_twin", get_twin)

    def patch_twin_reported_properties(self, reported_properties_patch):
        return super(IoTHubModuleClient, self).patch_twin_reported_properties(
            reported_properties_patch
        )

    patch_twin_reported_properties.__doc__ = (
        IoTHubModuleClient.patch_twin_reported_properties.__doc__
    )
    setattr(IoTHubModuleClient, "patch_twin_reported_properties", patch_twin_reported_properties)

    def receive_method_request(self, method_name=None, block=True, timeout=None):
        return super(IoTHubModuleClient, self).receive_method_request(method_name, block, timeout)

    receive_method_request.__doc__ = IoTHubModuleClient.receive_method_request.__doc__
    setattr(IoTHubModuleClient, "receive_method_request", receive_method_request)

    def receive_twin_desired_properties_patch(self, block=True, timeout=None):
        return super(IoTHubModuleClient, self).receive_twin_desired_properties_patch(block, timeout)

    receive_twin_desired_properties_patch.__doc__ = (
        IoTHubModuleClient.receive_twin_desired_properties_patch.__doc__
    )
    setattr(
        IoTHubModuleClient,
        "receive_twin_desired_properties_patch",
        receive_twin_desired_properties_patch,
    )

    def send_message(self, message):
        return super(IoTHubModuleClient, self).send_message(message)

    send_message.__doc__ = IoTHubModuleClient.send_message.__doc__
    setattr(IoTHubModuleClient, "send_message", send_message)

    def send_method_response(self, method_response):
        return super(IoTHubModuleClient, self).send_method_response(method_response)

    send_method_response.__doc__ = IoTHubModuleClient.send_method_response.__doc__
    setattr(IoTHubModuleClient, "send_method_response", send_method_response)

    def shutdown(self):
        return super(IoTHubModuleClient, self).shutdown()

    shutdown.__doc__ = IoTHubModuleClient.shutdown.__doc__
    setattr(IoTHubModuleClient, "shutdown", shutdown)

    def update_sastoken(self, sastoken):
        return super(IoTHubModuleClient, self).update_sastoken(sastoken)

    update_sastoken.__doc__ = IoTHubModuleClient.update_sastoken.__doc__
    setattr(IoTHubModuleClient, "update_sastoken", update_sastoken)

    def create_from_connection_string(cls, connection_string, **kwargs):
        return super(IoTHubModuleClient, cls).create_from_connection_string(
            connection_string, **kwargs
        )

    create_from_connection_string.__doc__ = IoTHubModuleClient.create_from_connection_string.__doc__
    setattr(
        IoTHubModuleClient,
        "create_from_connection_string",
        classmethod(create_from_connection_string),
    )

    def create_from_edge_environment(cls, **kwargs):
        return super(IoTHubModuleClient, cls).create_from_edge_environment(**kwargs)

    create_from_edge_environment.__doc__ = IoTHubModuleClient.create_from_edge_environment.__doc__
    setattr(
        IoTHubModuleClient,
        "create_from_edge_environment",
        classmethod(create_from_edge_environment),
    )

    def create_from_sastoken(cls, sastoken, **kwargs):
        return super(IoTHubModuleClient, cls).create_from_sastoken(sastoken, **kwargs)

    create_from_sastoken.__doc__ = IoTHubModuleClient.create_from_sastoken.__doc__
    setattr(IoTHubModuleClient, "create_from_sastoken", classmethod(create_from_sastoken))

    def create_from_x509_certificate(cls, x509, hostname, device_id, module_id, **kwargs):
        return super(IoTHubModuleClient, cls).create_from_x509_certificate(
            x509, hostname, device_id, module_id, **kwargs
        )

    create_from_x509_certificate.__doc__ = IoTHubModuleClient.create_from_x509_certificate.__doc__
    setattr(
        IoTHubModuleClient,
        "create_from_x509_certificate",
        classmethod(create_from_x509_certificate),
    )
    setattr(IoTHubModuleClient, "connected", IoTHubModuleClient.connected)
    setattr(
        IoTHubModuleClient, "on_background_exception", IoTHubModuleClient.on_background_exception
    )
    setattr(
        IoTHubModuleClient,
        "on_connection_state_change",
        IoTHubModuleClient.on_connection_state_change,
    )
    setattr(IoTHubModuleClient, "on_message_received", IoTHubModuleClient.on_message_received)
    setattr(
        IoTHubModuleClient,
        "on_method_request_received",
        IoTHubModuleClient.on_method_request_received,
    )
    setattr(
        IoTHubModuleClient, "on_new_sastoken_required", IoTHubModuleClient.on_new_sastoken_required
    )
    setattr(
        IoTHubModuleClient,
        "on_twin_desired_properties_patch_received",
        IoTHubModuleClient.on_twin_desired_properties_patch_received,
    )
    from azure.iot.device.provisioning.provisioning_device_client import (
        ProvisioningDeviceClient as ProvisioningDeviceClient,
    )

    def create_from_symmetric_key(
        cls, provisioning_host, registration_id, id_scope, symmetric_key, **kwargs
    ):
        return super(ProvisioningDeviceClient, cls).create_from_symmetric_key(
            provisioning_host, registration_id, id_scope, symmetric_key, **kwargs
        )

    create_from_symmetric_key.__doc__ = ProvisioningDeviceClient.create_from_symmetric_key.__doc__
    setattr(
        ProvisioningDeviceClient,
        "create_from_symmetric_key",
        classmethod(create_from_symmetric_key),
    )

    def create_from_x509_certificate(
        cls, provisioning_host, registration_id, id_scope, x509, **kwargs
    ):
        return super(ProvisioningDeviceClient, cls).create_from_x509_certificate(
            provisioning_host, registration_id, id_scope, x509, **kwargs
        )

    create_from_x509_certificate.__doc__ = (
        ProvisioningDeviceClient.create_from_x509_certificate.__doc__
    )
    setattr(
        ProvisioningDeviceClient,
        "create_from_x509_certificate",
        classmethod(create_from_x509_certificate),
    )
    setattr(
        ProvisioningDeviceClient,
        "provisioning_payload",
        ProvisioningDeviceClient.provisioning_payload,
    )



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
from azure.iot.device.common.evented_callback import EventedCallback
from azure.iot.device.common.pipeline import (
    pipeline_nucleus,
    pipeline_stages_base,
    pipeline_ops_base,
    pipeline_stages_http,
)

from azure.iot.device.iothub.pipeline import exceptions as pipeline_exceptions

from . import (
    pipeline_ops_iothub_http,
    pipeline_stages_iothub_http,
)

logger = logging.getLogger(__name__)


class HTTPPipeline(object):
    """Pipeline to communicate with Edge.
    Uses HTTP.
    """

    def __init__(self, pipeline_configuration):
        """
        Constructor for instantiating a pipeline adapter object.

        :param auth_provider: The authentication provider
        :param pipeline_configuration: The configuration generated based on user inputs
        """
        # NOTE: This pipeline DOES NOT handle SasToken management!
        # (i.e. using a SasTokenStage)
        # It instead relies on the parallel MQTT pipeline to handle that.
        #
        # Because they share a pipeline configuration, and MQTT has renewal logic we can be sure
        # that the SasToken in the pipeline configuration is valid.
        #
        # Furthermore, because HTTP doesn't require constant connections or long running tokens,
        # there's no need to reauthorize connections, so we can just pass the token from the config
        # when needed for auth.
        #
        # This is not an ideal solution, but it's the simplest one for the time being.

        # Contains data and information shared globally within the pipeline
        self._nucleus = pipeline_nucleus.PipelineNucleus(pipeline_configuration)

        self._pipeline = (
            pipeline_stages_base.PipelineRootStage(self._nucleus)
            .append_stage(pipeline_stages_iothub_http.IoTHubHTTPTranslationStage())
            .append_stage(pipeline_stages_http.HTTPTransportStage())
        )

        callback = EventedCallback()

        op = pipeline_ops_base.InitializePipelineOperation(callback=callback)

        self._pipeline.run_op(op)
        callback.wait_for_completion()

    def invoke_method(self, device_id, method_params, callback, module_id=None):
        """
        Send a request to the service to invoke a method on a target device or module.

        :param device_id: The target device id
        :param method_params: The method parameters to be invoked on the target client
        :param callback: callback which is called when request has been fulfilled.
            On success, this callback is called with the error=None.
            On failure, this callback is called with error set to the cause of the failure.
        :param module_id: The target module id

        The following exceptions are not "raised", but rather returned via the "error" parameter
            when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        """
        logger.debug("HTTPPipeline invoke_method called")
        if not self._nucleus.pipeline_configuration.method_invoke:
            # If this parameter is not set, that means that the pipeline was not generated by the edge environment. Method invoke only works for clients generated using the edge environment.
            error = pipeline_exceptions.PipelineRuntimeError(
                "invoke_method called, but it is only supported on module clients generated from an edge environment. If you are not using a module generated from an edge environment, you cannot use invoke_method"
            )
            return callback(error=error)

        def on_complete(op, error):
            callback(error=error, invoke_method_response=op.method_response)

        self._pipeline.run_op(
            pipeline_ops_iothub_http.MethodInvokeOperation(
                target_device_id=device_id,
                target_module_id=module_id,
                method_params=method_params,
                callback=on_complete,
            )
        )

    def get_storage_info_for_blob(self, blob_name, callback):
        """
        Sends a POST request to the IoT Hub service endpoint to retrieve an object that contains information for uploading via the Storage SDK.

        :param blob_name: The name of the blob that will be uploaded via the Azure Storage SDK.
        :param callback: callback which is called when request has been fulfilled.
            On success, this callback is called with the error=None, and the storage_info set to the information JSON received from the service.
            On failure, this callback is called with error set to the cause of the failure, and the storage_info=None.

        The following exceptions are not "raised", but rather returned via the "error" parameter
            when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        """
        logger.debug("HTTPPipeline get_storage_info_for_blob called")
        if not self._nucleus.pipeline_configuration.blob_upload:
            # If this parameter is not set, that means this is not a device client. Upload to blob is not supported on module clients.
            error = pipeline_exceptions.PipelineRuntimeError(
                "get_storage_info_for_blob called, but it is only supported for use with device clients. Ensure you are using a device client."
            )
            return callback(error=error)

        def on_complete(op, error):
            callback(error=error, storage_info=op.storage_info)

        self._pipeline.run_op(
            pipeline_ops_iothub_http.GetStorageInfoOperation(
                blob_name=blob_name, callback=on_complete
            )
        )

    def notify_blob_upload_status(
        self, correlation_id, is_success, status_code, status_description, callback
    ):
        """
        Sends a POST request to a IoT Hub service endpoint to notify the status of the Storage SDK call for a blob upload.

        :param str correlation_id: Provided by IoT Hub on get_storage_info_for_blob request.
        :param bool is_success: A boolean that indicates whether the file was uploaded successfully.
        :param int status_code: A numeric status code that is the status for the upload of the file to storage.
        :param str status_description: A description that corresponds to the status_code.

        :param callback: callback which is called when request has been fulfilled.
            On success, this callback is called with the error=None.
            On failure, this callback is called with error set to the cause of the failure.


        The following exceptions are not "raised", but rather returned via the "error" parameter
            when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        """
        logger.debug("HTTPPipeline notify_blob_upload_status called")
        if not self._nucleus.pipeline_configuration.blob_upload:
            # If this parameter is not set, that means this is not a device client. Upload to blob is not supported on module clients.
            error = pipeline_exceptions.PipelineRuntimeError(
                "notify_blob_upload_status called, but it is only supported for use with device clients. Ensure you are using a device client."
            )
            return callback(error=error)

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_iothub_http.NotifyBlobUploadStatusOperation(
                correlation_id=correlation_id,
                is_success=is_success,
                status_code=status_code,
                status_description=status_description,
                callback=on_complete,
            )
        )



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import abc
import time
import traceback
import uuid
import weakref
import threading
import queue
from . import pipeline_events_base
from . import pipeline_ops_base, pipeline_ops_mqtt
from . import pipeline_thread
from . import pipeline_exceptions
from .pipeline_nucleus import ConnectionState
from azure.iot.device.common import transport_exceptions, alarm
from azure.iot.device.common.auth import sastoken as st

logger = logging.getLogger(__name__)


class PipelineStage(abc.ABC):
    """
    Base class representing a stage in the processing pipeline.  Each stage is responsible for receiving
    PipelineOperation objects from the top, possibly processing them, and possibly passing them down.  It
    is also responsible for receiving PipelineEvent objects from the bottom, possibly processing them, and
    possibly passing them up.

    Each PipelineStage in the pipeline, is expected to act on some well-defined set of PipelineOperation
    types and/or some set of PipelineEvent types.  If any stage does not act on an operation or event, it
    should pass it to the next stage (for operations) or the previous stage (for events).  In this way, the
    pipeline implements the "chain of responsibility" design pattern (Gamma, et.al. "Design Patterns".
    Addison Wesley. 1995), with each stage being responsible for implementing some "rule" or "policy" of the
    pipeline, and each stage being ignorant of the stages that are before or after it in the pipeline.

    Each stage in the pipeline should act on the smallest set of rules possible, thus making stages small
    and easily testable.  Complex logic should be the exception and not the rule, and complex stages should
    operate on the most generic type of operation possible, thus allowing us to re-use complex logic for
    multiple cases.  The best way to do this is with "converter" stages that convert a specific operation to
    a more general one and with other converter stages that convert general operations to more specific ones.

    An example of a specific-to-generic stage is UseSkAuthProviderStage which takes a specific operation
    (use an auth provider) and converts it into something more generic (here is your device_id, etc, and use
    this SAS token when connecting).

    An example of a generic-to-specific stage is IoTHubMQTTTranslationStage which converts IoTHub operations
    (such as SendD2CMessageOperation) to MQTT operations (such as Publish).

    Each stage should also work in the broadest domain possible.  For example a generic stage (say
    "AutoConnectStage") that initiates a connection if any arbitrary operation needs a connection is more useful
    than having some MQTT-specific code that re-connects to the MQTT broker if the user calls Publish and
    there's no connection.

    One way to think about stages is to look at every "block of functionality" in your code and ask yourself
    "is this the one and only time I will need this code"?  If the answer is no, it might be worthwhile to
    implement that code in it's own stage in a very generic way.


    :ivar name: The name of the stage.  This is used primarily for logging
    :type name: str
    :ivar next: The next stage in the pipeline.  Set to None if this is the last stage in the pipeline.
    :type next: PipelineStage
    :ivar previous: The previous stage in the pipeline.  Set to None if this is the first stage in the pipeline.
    :type previous: PipelineStage
    :ivar nucleus: The pipeline's "nucleus" which contains global pipeline information, accessible
        from all stages
    :type nucleus: PipelineNucleus
    """

    def __init__(self):
        """
        Initializer for PipelineStage objects.
        """
        self.name = self.__class__.__name__
        self.next = None
        self.previous = None
        self.nucleus = None

    @pipeline_thread.runs_on_pipeline_thread
    def run_op(self, op):
        """
        Run the given operation.  This is the public function that outside callers would call to run an
        operation.  Derived classes should override the private _run_op function to implement
        stage-specific behavior.  When run_op returns, that doesn't mean that the operation has executed
        to completion.  Rather, it means that the pipeline has done something that will cause the
        operation to eventually execute to completion.  That might mean that something was sent over
        the network and some stage is waiting for a reply, or it might mean that the operation is sitting
        in a queue until something happens, or it could mean something entirely different.  The only
        thing you can assume is that the operation will _eventually_ complete successfully or fail, and the
        operation's callback will be called when that happens.

        :param PipelineOperation op: The operation to run.
        """
        try:
            self._run_op(op)
        except Exception as e:
            # This path is ONLY for unexpected errors. Expected errors should cause a fail completion
            # within ._run_op().
            #
            # We tag errors from here as logger.warning because, while we return them to the
            # caller and rely on the caller to handle them, they're somewhat unexpected and might be
            # worthy of investigation.

            # Do not use exc_info parameter on logger.* calls. This causes pytest to save the
            # traceback which saves stack frames which shows up as a leak
            logger.warning(msg="Unexpected error in {}._run_op() call".format(self))
            logger.warning(traceback.format_exc())

            # Only complete the operation if it is not already completed.
            # Attempting to complete a completed operation would raise an exception.
            if not op.completed:
                op.complete(error=e)
            else:
                # Note that this would be very unlikely to occur. It could only happen if a stage
                # was doing something after completing an operation, and an exception was raised,
                # which is unlikely because stages usually don't do anything after completing an
                # operation.
                raise e

    @pipeline_thread.runs_on_pipeline_thread
    def _run_op(self, op):
        """
        Implementation of the stage-specific function of .run_op(). Override this method instead of
        .run_op() in child classes in order to change how a stage behaves when running an operation.

        See the description of the .run_op() method for more discussion on what it means to "run"
        an operation.

        :param PipelineOperation op: The operation to run.
        """
        self.send_op_down(op)

    @pipeline_thread.runs_on_pipeline_thread
    def handle_pipeline_event(self, event):
        """
        Handle a pipeline event that arrives from the stage below this stage.  Derived
        classes should not override this function.  Any stage-specific handling of
        PipelineEvent objects should be implemented by overriding the private
        _handle_pipeline_event function in the derived stage.

        :param PipelineEvent event: The event that is being passed back up the pipeline
        """
        try:
            self._handle_pipeline_event(event)
        except Exception as e:
            # Do not use exc_info parameter on logger.* calls. This causes pytest to save the
            # traceback which saves stack frames which shows up as a leak
            logger.warning(
                msg="{}: Unexpected error in ._handle_pipeline_event() call: {}".format(self, e)
            )
            if self.previous:
                logger.warning("{}: Raising background exception")
                self.report_background_exception(e)
            else:
                # Nothing else we can do but log this. There exists no stage we can send the
                # exception to, and raising would send the error back down the pipeline.
                logger.warning(
                    "{}: Cannot report a background exception because there is no previous stage!"
                )

    @pipeline_thread.runs_on_pipeline_thread
    def _handle_pipeline_event(self, event):
        """
        Handle a pipeline event that arrives from the stage below this stage.  This
        is a function that is intended to be overridden in any stages that want to implement
        stage-specific handling of any events

        :param PipelineEvent event: The event that is being passed back up the pipeline
        """
        self.send_event_up(event)

    @pipeline_thread.runs_on_pipeline_thread
    def send_op_down(self, op):
        """
        Helper function to continue a given operation by passing it to the next stage
        in the pipeline.  If there is no next stage in the pipeline, this function
        will fail the operation and call complete_op to return the failure back up the
        pipeline.

        :param PipelineOperation op: Operation which is being passed on
        """
        if self.next:
            self.next.run_op(op)
        else:
            # This shouldn't happen if the pipeline was created correctly
            logger.warning(
                "{}({}): no next stage.cannot send op down. completing with error".format(
                    self.name, op.name
                )
            )
            raise pipeline_exceptions.PipelineRuntimeError(
                "{} not handled after {} stage with no next stage".format(op.name, self.name)
            )

    @pipeline_thread.runs_on_pipeline_thread
    def send_event_up(self, event):
        """
        Helper function to pass an event to the previous stage of the pipeline.  This is the default
        behavior of events while traveling through the pipeline. They start somewhere (maybe the
        bottom) and move up the pipeline until they're handled or until they error out.
        """
        if self.previous:
            self.previous.handle_pipeline_event(event)
        else:
            # This shouldn't happen if the pipeline was created correctly
            logger.critical(
                "{}({}): no previous stage. cannot send event up".format(event.name, self.name)
            )
            # NOTE: We can't report a background exception here because that involves
            # sending an event up, which is what got us into this problem in the first place.
            # Instead, raise, and let the method invoking this method handle it
            raise pipeline_exceptions.PipelineRuntimeError(
                "{} not handled after {} stage with no previous stage".format(event.name, self.name)
            )

    @pipeline_thread.runs_on_pipeline_thread
    def report_background_exception(self, e):
        """
        Send an exception up the pipeline that occurred in the background.
        These would typically be in response to unsolicited actions, such as receiving data or
        timer-based operations, which cannot be raised to the user because they occurred on a
        non-application thread.

        Note that this function leverages pipeline event flow, which means that any background
        exceptions in the core event flow itself become problematic (it's a good thing it's well
        tested then!)

        :param Exception e: The exception that occurred in the background
        """
        event = pipeline_events_base.BackgroundExceptionEvent(e)
        self.send_event_up(event)


class PipelineRootStage(PipelineStage):
    """
    Object representing the root of a pipeline.  This is where the functions to build
    the pipeline exist.  This is also where clients can add event handlers to receive
    events from the pipeline.

    :ivar on_pipeline_event_handler: Handler which can be set by users of the pipeline to
      receive PipelineEvent objects.  This is how users receive any "unsolicited"
      events from the pipeline (such as C2D messages).  This function is called with
      a PipelineEvent object every time any such event occurs.
    :type on_pipeline_event_handler: Function
    :ivar on_connected_handler: Handler which can be set by users of the pipeline to
      receive events every time the underlying transport connects
    :type on_connected_handler: Function
    :ivar on_disconnected_handler: Handler which can be set by users of the pipeline to
      receive events every time the underlying transport disconnects
    :type on_disconnected_handler: Function
    """

    def __init__(self, nucleus):
        super().__init__()
        self.on_pipeline_event_handler = None
        self.on_connected_handler = None
        self.on_disconnected_handler = None
        self.on_new_sastoken_required_handler = None
        self.on_background_exception_handler = None
        self.nucleus = nucleus

    def run_op(self, op):
        # CT-TODO: make this more elegant
        op.callback_stack[0] = pipeline_thread.invoke_on_callback_thread_nowait(
            op.callback_stack[0]
        )
        pipeline_thread.invoke_on_pipeline_thread(super().run_op)(op)

    def append_stage(self, new_stage):
        """
        Add the next stage to the end of the pipeline.  This is the function that callers
        use to build the pipeline by appending stages.  This function returns the root of
        the pipeline so that calls to this function can be chained together.

        :param PipelineStage new_stage: Stage to add to the end of the pipeline
        :returns: The root of the pipeline.
        """
        old_tail = self
        while old_tail.next:
            old_tail = old_tail.next
        old_tail.next = new_stage
        new_stage.previous = old_tail
        new_stage.nucleus = self.nucleus
        return self

    @pipeline_thread.runs_on_pipeline_thread
    def _handle_pipeline_event(self, event):
        """
        Override of the PipelineEvent handler.  Because this is the root of the pipeline,
        this function calls the on_pipeline_event_handler to pass the event to the
        caller.

        :param PipelineEvent event: Event to be handled, i.e. returned to the caller
          through the handle_pipeline_event (if provided).
        """
        # Base events that are common to all pipelines are handled here
        if isinstance(event, pipeline_events_base.ConnectedEvent):
            logger.debug(
                "{}: ConnectedEvent received. Calling on_connected_handler".format(self.name)
            )

            if self.on_connected_handler:
                pipeline_thread.invoke_on_callback_thread_nowait(self.on_connected_handler)()

        elif isinstance(event, pipeline_events_base.DisconnectedEvent):
            logger.debug(
                "{}: DisconnectedEvent received. Calling on_disconnected_handler".format(self.name)
            )
            if self.on_disconnected_handler:
                pipeline_thread.invoke_on_callback_thread_nowait(self.on_disconnected_handler)()

        elif isinstance(event, pipeline_events_base.NewSasTokenRequiredEvent):
            logger.debug(
                "{}: NewSasTokenRequiredEvent received. Calling on_new_sastoken_required_handler".format(
                    self.name
                )
            )
            if self.on_new_sastoken_required_handler:
                pipeline_thread.invoke_on_callback_thread_nowait(
                    self.on_new_sastoken_required_handler
                )()

        elif isinstance(event, pipeline_events_base.BackgroundExceptionEvent):
            logger.debug(
                "{}: BackgroundExceptionEvent received. Calling on_background_exception_handler".format(
                    self.name
                )
            )
            if self.on_background_exception_handler:
                pipeline_thread.invoke_on_callback_thread_nowait(
                    self.on_background_exception_handler
                )(event.e)

        # Events that are domain-specific and unique to each pipeline are handled by the provided
        # domain-specific .on_pipeline_event_handler
        else:
            if self.on_pipeline_event_handler:
                pipeline_thread.invoke_on_callback_thread_nowait(self.on_pipeline_event_handler)(
                    event
                )
            else:
                # unexpected condition: we should be handling all pipeline events
                logger.debug("incoming {} event with no handler.  dropping.".format(event.name))


# NOTE: This stage could be a candidate for being refactored into some kind of other
# pipeline-related structure. What's odd about it as a stage is that it doesn't really respond
# to operations or events so much as it spawns them on a timer.
# Perhaps some kind of... Pipeline Daemon?
class SasTokenStage(PipelineStage):
    # Amount of time, in seconds, prior to token expiration to trigger alarm
    DEFAULT_TOKEN_UPDATE_MARGIN = 120

    def __init__(self):
        super().__init__()
        # Indicates when token needs to be updated
        self._token_update_alarm = None
        # Indicates when to retry a failed reauthorization attempt
        # (only used with renewable SAS auth)
        self._reauth_retry_timer = None

    @pipeline_thread.runs_on_pipeline_thread
    def _run_op(self, op):
        if (
            isinstance(op, pipeline_ops_base.InitializePipelineOperation)
            and self.nucleus.pipeline_configuration.sastoken is not None
        ):
            # Start an alarm (renewal or replacement depending on token type)
            self._start_token_update_alarm()
            self.send_op_down(op)
        elif (
            isinstance(op, pipeline_ops_base.ReauthorizeConnectionOperation)
            and self.nucleus.pipeline_configuration.sastoken is not None
        ):
            # NOTE 1: This case (currently) implies that we are using Non-Renewable SAS,
            # although it's not enforced here (it's a product of how the pipeline and client are
            # configured overall)

            # NOTE 2: There's a theoretically possible case where the new token has the same expiry
            # time as the old token, and thus a new update alarm wouldn't really be required, but
            # I don't want to include the complexity of checking. Just start a new alarm anyway.

            # NOTE 3: Yeah, this is the same logic as the above case for the InitializePipeline op,
            # but if it weren't separate, how would you get all these nice informative comments?
            # (Also, it leaves room for the logic to change in the future)
            self._start_token_update_alarm()
            self.send_op_down(op)
        elif isinstance(op, pipeline_ops_base.ShutdownPipelineOperation):
            self._cancel_token_update_alarm()
            self._cancel_reauth_retry_timer()
            self.send_op_down(op)
        else:
            self.send_op_down(op)

    @pipeline_thread.runs_on_pipeline_thread
    def _cancel_token_update_alarm(self):
        """Cancel and delete any pending update alarm"""
        old_alarm = self._token_update_alarm
        self._token_update_alarm = None
        if old_alarm:
            logger.debug("Cancelling SAS Token update alarm")
            old_alarm.cancel()
            old_alarm = None

    @pipeline_thread.runs_on_pipeline_thread
    def _cancel_reauth_retry_timer(self):
        """Cancel and delete any pending reauth retry timer"""
        old_reauth_retry_timer = self._reauth_retry_timer
        self._reauth_retry_timer = None
        if old_reauth_retry_timer:
            logger.debug("Cancelling reauthorization retry timer")
            old_reauth_retry_timer.cancel()
            old_reauth_retry_timer = None

    @pipeline_thread.runs_on_pipeline_thread
    def _start_token_update_alarm(self):
        """Begin an update alarm.
        If using a RenewableSasToken, when the alarm expires the token will be automatically
        renewed, and a new alarm will be set.

        If using a NonRenewableSasToken, when the alarm expires, it will trigger a
        NewSasTokenRequiredEvent to signal that a new SasToken must be manually provided.
        """
        self._cancel_token_update_alarm()

        update_time = (
            self.nucleus.pipeline_configuration.sastoken.expiry_time
            - self.DEFAULT_TOKEN_UPDATE_MARGIN
        )

        # On Windows platforms, the threading event TIMEOUT_MAX (approximately 49.7 days) could
        # conceivably be less than the SAS lifespan, which means we may need to update the token
        # before the lifespan ends.
        # If we really wanted to adjust this in the future to use the entire SAS lifespan, we could
        # implement Alarms that trigger other Alarms, but for now, just forcing a token update
        # is good enough.
        # Note that this doesn't apply to (most) Unix platforms, where TIMEOUT_MAX is 292.5 years.
        if (update_time - time.time()) > threading.TIMEOUT_MAX:
            update_time = time.time() + threading.TIMEOUT_MAX
            logger.warning(
                "SAS Token expiration ({expiry} seconds) exceeds max scheduled renewal time ({max} seconds). Will be renewing after {max} seconds instead".format(
                    expiry=self.nucleus.pipeline_configuration.sastoken.expiry_time,
                    max=threading.TIMEOUT_MAX,
                )
            )

        self_weakref = weakref.ref(self)

        # For renewable SasTokens, create an alarm that will automatically renew the token,
        # and then start another alarm.
        if isinstance(self.nucleus.pipeline_configuration.sastoken, st.RenewableSasToken):
            logger.debug(
                "{}: Scheduling automatic SAS Token renewal at epoch time: {}".format(
                    self.name, update_time
                )
            )

            @pipeline_thread.invoke_on_pipeline_thread_nowait
            def renew_token():
                this = self_weakref()
                # Cancel any token reauth retry timer in progress (from a previous renewal)
                this._cancel_reauth_retry_timer()
                logger.info("{}: Renewing SAS Token...".format(self.name))
                # Renew the token
                sastoken = this.nucleus.pipeline_configuration.sastoken
                try:
                    sastoken.refresh()
                except st.SasTokenError as e:
                    logger.error("{}: SAS Token renewal failed".format(self.name))
                    this.report_background_exception(e)
                    # TODO: then what? How do we respond to this? Retry?
                    # What if it never works and the token expires?
                else:
                    # If the pipeline is already connected, send order to reauthorize the connection
                    # now that token has been renewed. If the pipeline is not currently connected,
                    # there is no need to do this, as the next connection will be using the new
                    # credentials.
                    if this.nucleus.connected:
                        this._reauthorize()

                    # Once again, start a renewal alarm
                    this._start_token_update_alarm()

            self._token_update_alarm = alarm.Alarm(update_time, renew_token)

        # For nonrenewable SasTokens, create an alarm that will issue a NewSasTokenRequiredEvent
        else:
            logger.debug(
                "Scheduling manual SAS Token renewal at epoch time: {}".format(update_time)
            )

            @pipeline_thread.invoke_on_pipeline_thread_nowait
            def request_new_token():
                this = self_weakref()
                logger.info("Requesting new SAS Token....")
                # Send request
                this.send_event_up(pipeline_events_base.NewSasTokenRequiredEvent())

            self._token_update_alarm = alarm.Alarm(update_time, request_new_token)

        self._token_update_alarm.daemon = True
        self._token_update_alarm.start()

    @pipeline_thread.runs_on_pipeline_thread
    def _reauthorize(self):
        self_weakref = weakref.ref(self)

        @pipeline_thread.runs_on_pipeline_thread
        def on_reauthorize_complete(op, error):
            this = self_weakref()
            if error:
                logger.info(
                    "{}: Connection reauthorization failed.  Error={}".format(this.name, error)
                )
                self.report_background_exception(error)
                # If connection has not been somehow re-established, we need to keep trying
                # because for the reauthorization to originally have been issued, we were in
                # a connected state.
                # NOTE: we only do this if connection retry is enabled on the pipeline. If it is,
                # we have a contract to maintain a connection. If it has been disabled, we have
                # a contract to not do so.
                # NOTE: We can't rely on the ConnectionStateStage to do this because 1) the pipeline
                # stages should stand on their own, and 2) if the reauth failed, the ConnectionStateStage
                # wouldn't know to reconnect, because the expected state of a failed reauth is
                # to be disconnected.
                if (
                    not this.nucleus.connected
                    and this.nucleus.pipeline_configuration.connection_retry
                ):
                    logger.info("{}: Retrying connection reauthorization".format(this.name))
                    # No need to cancel the timer, because if this is running, it has already ended

                    @pipeline_thread.invoke_on_pipeline_thread_nowait
                    def retry_reauthorize():
                        # We need to check this when the timer expires as well as before creating
                        # the timer in case connection has been re-established while timer was
                        # running
                        if not this.nucleus.connected:
                            this._reauthorize()

                    this._reauth_retry_timer = threading.Timer(
                        this.nucleus.pipeline_configuration.connection_retry_interval,
                        retry_reauthorize,
                    )
                    this._reauth_retry_timer.daemon = True
                    this._reauth_retry_timer.start()

            else:
                logger.info("{}: Connection reauthorization successful".format(this.name))

        logger.info("{}: Starting reauthorization process for new SAS token".format(self.name))
        self.send_op_down(
            pipeline_ops_base.ReauthorizeConnectionOperation(callback=on_reauthorize_complete)
        )


class AutoConnectStage(PipelineStage):
    """
    This stage is responsible for ensuring that the protocol is connected when
    it needs to be connected.
    """

    @pipeline_thread.runs_on_pipeline_thread
    def _run_op(self, op):
        # Any operation that requires a connection can trigger a connection if
        # we're not connected and the auto-connect feature is enabled.
        if (
            op.needs_connection
            and not self.nucleus.connected
            and self.nucleus.pipeline_configuration.auto_connect
        ):
            logger.debug(
                "{}({}): Op needs connection.  Queueing this op and starting a ConnectionOperation".format(
                    self.name, op.name
                )
            )
            self._do_connect(op)

        else:
            self.send_op_down(op)

    @pipeline_thread.runs_on_pipeline_thread
    def _do_connect(self, op):
        """
        Start connecting the transport in response to some operation
        """
        # Alias to avoid overload within the callback below
        # CT-TODO: remove the need for this with better callback semantics
        op_needs_connect = op

        # function that gets called after we're connected.
        @pipeline_thread.runs_on_pipeline_thread
        def on_connect_op_complete(op, error):
            if error:
                logger.debug(
                    "{}({}): Connection failed.  Completing with failure because of connection failure: {}".format(
                        self.name, op_needs_connect.name, error
                    )
                )
                op_needs_connect.complete(error=error)
            else:
                logger.debug(
                    "{}({}): connection is complete.  Running op that triggered connection.".format(
                        self.name, op_needs_connect.name
                    )
                )
                self.run_op(op_needs_connect)

        # call down to the next stage to connect.
        logger.debug("{}({}): calling down with Connect operation".format(self.name, op.name))
        self.send_op_down(pipeline_ops_base.ConnectOperation(callback=on_connect_op_complete))


class CoordinateRequestAndResponseStage(PipelineStage):
    """
    Pipeline stage which is responsible for coordinating RequestAndResponseOperation operations.  For each
    RequestAndResponseOperation operation, this stage passes down a RequestOperation operation and waits for
    an ResponseEvent event.  All other events are passed down unmodified.
    """

    def __init__(self):
        super().__init__()
        self.pending_responses = {}

    @pipeline_thread.runs_on_pipeline_thread
    def _run_op(self, op):
        if isinstance(op, pipeline_ops_base.RequestAndResponseOperation):
            # Convert RequestAndResponseOperation operation into a RequestOperation operation
            # and send it down.  A lower level will convert the RequestOperation into an
            # actual protocol client operation.  The RequestAndResponseOperation operation will be
            # completed when the corresponding IotResponse event is received in this stage.

            request_id = str(uuid.uuid4())

            logger.debug(
                "{}({}): adding request {} to pending list".format(self.name, op.name, request_id)
            )
            self.pending_responses[request_id] = op

            self._send_request_down(request_id, op)

        else:
            self.send_op_down(op)

    @pipeline_thread.runs_on_pipeline_thread
    def _send_request_down(self, request_id, op):
        # Alias to avoid overload within the callback below
        # CT-TODO: remove the need for this with better callback semantics
        op_waiting_for_response = op

        @pipeline_thread.runs_on_pipeline_thread
        def on_send_request_done(op, error):
            logger.debug(
                "{}({}): Finished sending {} request to {} resource {}".format(
                    self.name,
                    op_waiting_for_response.name,
                    op_waiting_for_response.request_type,
                    op_waiting_for_response.method,
                    op_waiting_for_response.resource_location,
                )
            )
            if error:
                logger.debug(
                    "{}({}): removing request {} from pending list".format(
                        self.name, op_waiting_for_response.name, request_id
                    )
                )
                # if there's no pending response for the given request_id, there's nothing to delete
                if request_id in self.pending_responses:
                    del self.pending_responses[request_id]
                    op_waiting_for_response.complete(error=error)
                else:
                    # NOTE: This shouldn't ever happen under normal conditions, but the following logic
                    # ensures that, if it does, it's handled safely.
                    logger.debug(
                        "{}({}): request_id {} not found in pending list.  Unexpected behavior.  Dropping".format(
                            self.name, op_waiting_for_response.name, request_id
                        )
                    )
                    pass
            else:
                # request sent.  Nothing to do except wait for the response
                pass

        logger.debug(
            "{}({}): Sending {} request to {} resource {}".format(
                self.name, op.name, op.request_type, op.method, op.resource_location
            )
        )

        new_op = pipeline_ops_base.RequestOperation(
            method=op.method,
            resource_location=op.resource_location,
            request_body=op.request_body,
            request_id=request_id,
            request_type=op.request_type,
            callback=on_send_request_done,
            query_params=op.query_params,
        )
        self.send_op_down(new_op)

    @pipeline_thread.runs_on_pipeline_thread
    def _handle_pipeline_event(self, event):
        if isinstance(event, pipeline_events_base.ResponseEvent):
            # match ResponseEvent events to the saved dictionary of RequestAndResponseOperation
            # operations which have not received responses yet.  If the operation is found,
            # complete it.

            logger.debug(
                "{}({}): Handling event with request_id {}".format(
                    self.name, event.name, event.request_id
                )
            )
            if event.request_id in self.pending_responses:
                op = self.pending_responses[event.request_id]
                del self.pending_responses[event.request_id]
                op.status_code = event.status_code
                op.response_body = event.response_body
                op.retry_after = event.retry_after
                logger.debug(
                    "{}({}): Completing {} request to {} resource {} with status {}".format(
                        self.name,
                        op.name,
                        op.request_type,
                        op.method,
                        op.resource_location,
                        op.status_code,
                    )
                )
                op.complete()
            else:
                logger.info(
                    "{}({}): request_id {} not found in pending list.  Nothing to do.  Dropping".format(
                        self.name, event.name, event.request_id
                    )
                )

        else:
            self.send_event_up(event)


class OpTimeoutStage(PipelineStage):
    """
    The purpose of the timeout stage is to add timeout errors to select operations

    The timeout_intervals attribute contains a list of operations to track along with
    their timeout values.  Right now this list is hard-coded but the operations and
    intervals will eventually become a parameter.

    For each operation that needs a timeout check, this stage will add a timer to
    the operation.  If the timer elapses, this stage will fail the operation with
    a OperationTimeout.  The intention is that a higher stage will know what to
    do with that error and act accordingly (either return the error to the user or
    retry).

    This stage currently assumes that all timed out operation are just "lost".
    It does not attempt to cancel the operation, as Paho doesn't have a way to
    cancel an operation, and with QOS=1, sending a pub or sub twice is not
    catastrophic.

    Also, as a long-term plan, the operations that need to be watched for timeout
    will become an initialization parameter for this stage so that different
    instances of this stage can watch for timeouts on different operations.
    This will be done because we want a lower-level timeout stage which can watch
    for timeouts at the MQTT level, and we want a higher-level timeout stage which
    can watch for timeouts at the iothub level.  In this way, an MQTT operation that
    times out can be retried as an MQTT operation and a higher-level IoTHub operation
    which times out can be retried as an IoTHub operation (which might necessitate
    redoing multiple MQTT operations).
    """

    def __init__(self):
        super().__init__()
        # use a fixed list and fixed intervals for now.  Later, this info will come in
        # as an init param or a retry policy
        self.timeout_intervals = {
            pipeline_ops_mqtt.MQTTSubscribeOperation: 10,
            pipeline_ops_mqtt.MQTTUnsubscribeOperation: 10,
            # Only Sub and Unsub are here because MQTT auto retries pub
        }

    @pipeline_thread.runs_on_pipeline_thread
    def _run_op(self, op):
        if type(op) in self.timeout_intervals:
            # Create a timer to watch for operation timeout on this op and attach it
            # to the op.
            self_weakref = weakref.ref(self)

            @pipeline_thread.invoke_on_pipeline_thread_nowait
            def on_timeout():
                this = self_weakref()
                logger.info("{}({}): returning timeout error".format(this.name, op.name))
                op.complete(
                    error=pipeline_exceptions.OperationTimeout(
                        "operation timed out before protocol client could respond"
                    )
                )

            logger.debug("{}({}): Creating timer".format(self.name, op.name))
            op.timeout_timer = threading.Timer(self.timeout_intervals[type(op)], on_timeout)
            op.timeout_timer.start()

            # Send the op down, but intercept the return of the op so we can
            # remove the timer when the op is done
            op.add_callback(self._clear_timer)
            logger.debug("{}({}): Sending down".format(self.name, op.name))
            self.send_op_down(op)
        else:
            self.send_op_down(op)

    @pipeline_thread.runs_on_pipeline_thread
    def _clear_timer(self, op, error):
        # When an op comes back, delete the timer and pass it right up.
        if op.timeout_timer:
            logger.debug("{}({}): Cancelling timer".format(self.name, op.name))
            op.timeout_timer.cancel()
            op.timeout_timer = None


class RetryStage(PipelineStage):
    """
    The purpose of the retry stage is to watch specific operations for specific
    errors and retry the operations as appropriate.

    Unlike the OpTimeoutStage, this stage will never need to worry about cancelling
    failed operations.  When an operation is retried at this stage, it is already
    considered "failed", so no cancellation needs to be done.
    """

    def __init__(self):
        super().__init__()
        # Retry intervals are hardcoded for now. Later, they come in as an
        # init param, probably via retry policy.
        self.retry_intervals = {
            pipeline_ops_mqtt.MQTTSubscribeOperation: 20,
            pipeline_ops_mqtt.MQTTUnsubscribeOperation: 20,
            # Only Sub and Unsub are here because MQTT auto retries pub
        }
        self.ops_waiting_to_retry = []

    @pipeline_thread.runs_on_pipeline_thread
    def _run_op(self, op):
        """
        Send all ops down and intercept their return to "watch for retry"
        """
        if self._should_watch_for_retry(op):
            op.add_callback(self._do_retry_if_necessary)
            self.send_op_down(op)
        else:
            self.send_op_down(op)

    @pipeline_thread.runs_on_pipeline_thread
    def _should_watch_for_retry(self, op):
        """
        Return True if this op needs to be watched for retry.  This can be
        called before the op runs.
        """
        return type(op) in self.retry_intervals

    @pipeline_thread.runs_on_pipeline_thread
    def _should_retry(self, op, error):
        """
        Return True if this op needs to be retried.  This must be called after
        the op completes.
        """
        if error:
            if self._should_watch_for_retry(op):
                if isinstance(error, pipeline_exceptions.OperationTimeout):
                    return True
        return False

    @pipeline_thread.runs_on_pipeline_thread
    def _do_retry_if_necessary(self, op, error):
        """
        Handler which gets called when operations are complete.  This function
        is where we check to see if a retry is necessary and set a "retry timer"
        which can be used to send the op down again.
        """
        if self._should_retry(op, error):
            self_weakref = weakref.ref(self)

            @pipeline_thread.invoke_on_pipeline_thread_nowait
            def do_retry():
                this = self_weakref()
                logger.debug("{}({}): retrying".format(this.name, op.name))
                op.retry_timer.cancel()
                op.retry_timer = None
                this.ops_waiting_to_retry.remove(op)
                # Don't just send it down directly.  Instead, go through run_op so we get
                # retry functionality this time too
                this.run_op(op)

            interval = self.retry_intervals[type(op)]
            logger.info(
                "{}({}): Op needs retry with interval {} because of {}.  Setting timer.".format(
                    self.name, op.name, interval, error
                )
            )

            # if we don't keep track of this op, it might get collected.
            op.halt_completion()
            self.ops_waiting_to_retry.append(op)
            op.retry_timer = threading.Timer(self.retry_intervals[type(op)], do_retry)
            op.retry_timer.start()

        else:
            if op.retry_timer:
                op.retry_timer.cancel()
                op.retry_timer = None


class ConnectionStateStage(PipelineStage):

    intermediate_states = [
        ConnectionState.CONNECTING,
        ConnectionState.DISCONNECTING,
        ConnectionState.REAUTHORIZING,
    ]
    transient_connect_errors = [
        pipeline_exceptions.OperationCancelled,
        pipeline_exceptions.OperationTimeout,
        pipeline_exceptions.OperationError,
        transport_exceptions.ConnectionFailedError,
        transport_exceptions.ConnectionDroppedError,
        transport_exceptions.TlsExchangeAuthError,
    ]

    def __init__(self):
        super().__init__()
        self.reconnect_timer = None
        self.waiting_ops = queue.Queue()

        # NOTE: In this stage states are both checked, and changed, but there is no lock to protect
        # this state value, or the logic that surrounds it from multithreading. This is because due
        # to the threading model of the pipeline, there is a dedicated pipeline thread that handles
        # everything that runs here, and it can only be doing one thing at a time. Thus we don't
        # need to have a threading lock on our state, or be concerned with how atomic things are.

    @pipeline_thread.runs_on_pipeline_thread
    def _run_op(self, op):

        # If receiving an operation while the connection state is changing, wait for the
        # connection state to reach a stable state before continuing.
        if self.nucleus.connection_state in self.intermediate_states:
            logger.debug(
                "{}({}): State is {} - waiting for in-progress operation to finish".format(
                    self.name, op.name, self.nucleus.connection_state
                )
            )
            self.waiting_ops.put_nowait(op)

        else:
            if isinstance(op, pipeline_ops_base.ConnectOperation):
                if self.nucleus.connection_state is ConnectionState.CONNECTED:
                    logger.debug(
                        "{}({}): State is already CONNECTED. Completing operation".format(
                            self.name, op.name
                        )
                    )
                    op.complete()
                elif self.nucleus.connection_state is ConnectionState.DISCONNECTED:
                    logger.debug(
                        "{}({}): State changes DISCONNECTED -> CONNECTING. Sending op down".format(
                            self.name, op.name
                        )
                    )
                    self.nucleus.connection_state = ConnectionState.CONNECTING
                    self._add_connection_op_callback(op)
                    self.send_op_down(op)
                else:
                    # This should be impossible to reach. If the state were intermediate, it
                    # would have been added to the waiting ops queue above.
                    logger.warning(
                        "{}({}): Invalid State - {}".format(
                            self.name, op.name, self.nucleus.connection_state
                        )
                    )
                    self.send_op_down(op)

            elif isinstance(op, pipeline_ops_base.DisconnectOperation):
                # First, always clear any reconnect timer. Because a manual disconnection is
                # occurring, we won't want to be reconnecting any more.
                self._clear_reconnect_timer()

                if self.nucleus.connection_state is ConnectionState.CONNECTED:
                    logger.debug(
                        "{}({}): State changes CONNECTED -> DISCONNECTING. Sending op down.".format(
                            self.name, op.name
                        )
                    )
                    self.nucleus.connection_state = ConnectionState.DISCONNECTING
                    self._add_connection_op_callback(op)
                    self.send_op_down(op)
                elif self.nucleus.connection_state is ConnectionState.DISCONNECTED:
                    logger.debug(
                        "{}({}): State is already DISCONNECTED. Completing operation".format(
                            self.name, op.name
                        )
                    )
                    op.complete()
                else:
                    # This should be impossible to reach. If the state were intermediate, it
                    # would have been added to the waiting ops queue above.
                    logger.warning(
                        "{}({}): Invalid State - {}".format(
                            self.name, op.name, self.nucleus.connection_state
                        )
                    )
                    self.send_op_down(op)

            elif isinstance(op, pipeline_ops_base.ReauthorizeConnectionOperation):
                if self.nucleus.connection_state is ConnectionState.CONNECTED:
                    logger.debug(
                        "{}({}): State changes CONNECTED -> REAUTHORIZING. Sending op down.".format(
                            self.name, op.name
                        )
                    )
                    self.nucleus.connection_state = ConnectionState.REAUTHORIZING
                    self._add_connection_op_callback(op)
                    self.send_op_down(op)
                elif self.nucleus.connection_state is ConnectionState.DISCONNECTED:
                    logger.debug(
                        "{}({}): State changes DISCONNECTED -> REAUTHORIZING. Sending op down".format(
                            self.name, op.name
                        )
                    )
                    self.nucleus.connection_state = ConnectionState.REAUTHORIZING
                    self._add_connection_op_callback(op)
                    self.send_op_down(op)
                else:
                    # This should be impossible to reach. If the state were intermediate, it
                    # would have been added to the waiting ops queue above.
                    logger.warning(
                        "{}({}): Invalid State - {}".format(
                            self.name, op.name, self.nucleus.connection_state
                        )
                    )
                    self.send_op_down(op)

            elif isinstance(op, pipeline_ops_base.ShutdownPipelineOperation):
                self._clear_reconnect_timer()
                # Cancel all pending ops so they don't hang
                while not self.waiting_ops.empty():
                    waiting_op = self.waiting_ops.get_nowait()
                    cancel_error = pipeline_exceptions.OperationCancelled(
                        "Operation waiting in ConnectionStateStage cancelled by shutdown"
                    )
                    waiting_op.complete(error=cancel_error)
                self.send_op_down(op)

            else:
                self.send_op_down(op)

    @pipeline_thread.runs_on_pipeline_thread
    def _handle_pipeline_event(self, event):

        if isinstance(event, pipeline_events_base.ConnectedEvent):
            # First, clear the reconnect timer no matter what.
            # We are now connected, so any ongoing reconnect is unnecessary
            self._clear_reconnect_timer()

            # EXPECTED CONNECTION (ConnectOperation was previously issued)
            if self.nucleus.connection_state is ConnectionState.CONNECTING:
                logger.debug(
                    "{}({}): State changes CONNECTING -> CONNECTED. Connection established".format(
                        self.name, event.name
                    )
                )
                self.nucleus.connection_state = ConnectionState.CONNECTED

            # EXPECTED CONNECTION (ReauthorizeConnectionOperation was previously issued)
            elif self.nucleus.connection_state is ConnectionState.REAUTHORIZING:
                logger.debug(
                    "{}({}): State changes REAUTHORIZING -> CONNECTED. Connection re-established after re-authentication".format(
                        self.name, event.name
                    )
                )
                self.nucleus.connection_state = ConnectionState.CONNECTED

            # BAD STATE (this block should not be reached)
            else:
                logger.warning(
                    "{}: ConnectedEvent received while in unexpected state - {}".format(
                        self.name, self.nucleus.connection_state
                    )
                )
                logger.debug(
                    "{}({}): State changes {} -> CONNECTED. Unexpected connection".format(
                        self.name, event.name, self.nucleus.connection_state
                    )
                )
                self.nucleus.connection_state = ConnectionState.CONNECTED

        elif isinstance(event, pipeline_events_base.DisconnectedEvent):
            # UNEXPECTED DISCONNECTION (i.e. Connection has been lost)
            if self.nucleus.connection_state is ConnectionState.CONNECTED:

                # Set the state change before starting the timer in order to make sure
                # there's no issues when the timer expires. The pipeline threading model should
                # already be preventing any weirdness with timing, but can't hurt to do this
                # as well.
                self.nucleus.connection_state = ConnectionState.DISCONNECTED

                if self.nucleus.pipeline_configuration.connection_retry:
                    # When we get disconnected, we try to reconnect as soon as we can. We set a
                    # timer here that will start the process in another thread because we don't
                    # want to hold up the event flow
                    logger.debug(
                        "{}({}): State changes CONNECTED -> DISCONNECTED. Attempting to reconnect".format(
                            self.name, event.name
                        )
                    )
                    self._start_reconnect_timer(0.01)
                else:
                    logger.debug(
                        "{}({}): State changes CONNECTED -> DISCONNECTED. Not attempting to reconnect (Connection retry disabled)".format(
                            self.name, event.name
                        )
                    )

            # EXPECTED DISCONNECTION (DisconnectOperation was previously issued)
            elif self.nucleus.connection_state is ConnectionState.DISCONNECTING:
                # No reconnect timer will be created.
                logger.debug(
                    "{}({}): State changes DISCONNECTING -> DISCONNECTED. Not attempting to reconnect (User-initiated disconnect)".format(
                        self.name, event.name
                    )
                )
                self.nucleus.connection_state = ConnectionState.DISCONNECTED

            # EXPECTED DISCONNECTION (Reauthorization process)
            elif self.nucleus.connection_state is ConnectionState.REAUTHORIZING:
                # ConnectionState will remain REAUTHORIZING until completion of the process
                # upon re-establishing the connection

                # NOTE: There is a ~small~ chance of a false positive here if an unexpected
                # disconnection occurs while a ReauthorizationOperation is in flight.
                # However, it will sort itself out - the ensuing connect that occurs as part
                # of the reauthorization will restore connection (no harm done) or it will
                # fail, at which point the failure was a result of a manual operation and
                # reconnection is not supposed to occur. So either way, we end up where we want
                # to be despite the false positive - just be aware that this can happen.
                logger.debug(
                    "{}({}): Not attempting to reconnect (Reauthorization in progress)".format(
                        self.name, event.name
                    )
                )

            # BAD STATE (this block should not be reached)
            else:
                logger.warning(
                    "{}: DisconnectEvent received while in unexpected state - {}".format(
                        self.name, self.nucleus.connection_state
                    )
                )
                logger.debug(
                    "{}({}): State changes {} -> DISCONNECTED. Unexpected disconnect in unexpected state".format(
                        self.name, event.name, self.nucleus.connection_state
                    )
                )
                self.nucleus.connection_state = ConnectionState.DISCONNECTED

        # In all cases the event is sent up
        self.send_event_up(event)

    @pipeline_thread.runs_on_pipeline_thread
    def _add_connection_op_callback(self, op):
        """Adds callback to a connection op passing through to do necessary stage upkeep"""
        self_weakref = weakref.ref(self)

        @pipeline_thread.runs_on_pipeline_thread
        def on_complete(op, error):
            this = self_weakref()
            # If error, set us back to a DISCONNECTED state. It doesn't matter what kind of
            # connection op this was, any failure should result in a disconnected state.

            # NOTE: Due to the stage waiting any ops if an ongoing connection op is in-progress
            # as well as the way that the reconnection process checks if there is an in-progress
            # connection op (and punts the reconnect if so), there is no risk here of setting
            # directly to DISCONNECTED - the intermediate state being overwritten is always going
            # to be due to this op that is now completing, we can be assured of that.
            if error:
                logger.debug(
                    "{}({}): failed, state change {} -> DISCONNECTED".format(
                        this.name, op.name, this.nucleus.connection_state
                    )
                )
                this.nucleus.connection_state = ConnectionState.DISCONNECTED

            # Allow the next waiting op to proceed (if any)
            this._run_all_waiting_ops()

        op.add_callback(on_complete)

    @pipeline_thread.runs_on_pipeline_thread
    def _run_all_waiting_ops(self):

        if not self.waiting_ops.empty():
            queuecopy = self.waiting_ops
            self.waiting_ops = queue.Queue()

            while not queuecopy.empty():
                next_op = queuecopy.get_nowait()
                if not next_op.completed:
                    logger.debug(
                        "{}: Resolving next waiting op: {}".format(self.name, next_op.name)
                    )
                    self.run_op(next_op)

    @pipeline_thread.runs_on_pipeline_thread
    def _reconnect(self):
        self_weakref = weakref.ref(self)

        @pipeline_thread.runs_on_pipeline_thread
        def on_reconnect_complete(op, error):
            this = self_weakref()
            if this:
                logger.debug(
                    "{}({}): on_connect_complete error={} state={} ".format(
                        this.name,
                        op.name,
                        error,
                        this.nucleus.connection_state,
                    )
                )

                if error:
                    # Set state back to DISCONNECTED so as not to block anything else
                    logger.debug(
                        "{}: State change {} -> DISCONNECTED".format(
                            this.name, this.nucleus.connection_state
                        )
                    )
                    this.nucleus.connection_state = ConnectionState.DISCONNECTED

                    # report background exception to indicate this failure occurred
                    this.report_background_exception(error)

                    # Determine if should try reconnect again
                    if this._should_reconnect(error):
                        # transient errors can cause a reconnect attempt
                        logger.debug(
                            "{}: Reconnect failed. Starting reconnection timer".format(this.name)
                        )
                        this._start_reconnect_timer(
                            this.nucleus.pipeline_configuration.connection_retry_interval
                        )
                    else:
                        # all others are permanent errors
                        logger.debug(
                            "{}: Cannot reconnect. Ending reconnection process".format(this.name)
                        )

                # Now see if there's anything that may have blocked waiting for us to finish
                this._run_all_waiting_ops()

        # NOTE: I had considered leveraging the run_op infrastructure instead of sending this
        # directly down. Ultimately however, I think it's best to keep reconnects completely
        # distinct from other operations that come through the pipeline - for instance, we don't
        # really want them to end up queued up behind other operations via the .waiting_ops queue.
        # Reconnects have a top priority.
        op = pipeline_ops_base.ConnectOperation(callback=on_reconnect_complete)
        self.send_op_down(op)

    @pipeline_thread.runs_on_pipeline_thread
    def _should_reconnect(self, error):
        """Returns True if a reconnect should occur in response to an error, False otherwise"""
        if self.nucleus.pipeline_configuration.connection_retry:
            if type(error) in self.transient_connect_errors:
                return True
        return False

    @pipeline_thread.runs_on_pipeline_thread
    def _start_reconnect_timer(self, delay):
        """
        Set a timer to reconnect after some period of time
        """
        self._clear_reconnect_timer()

        self_weakref = weakref.ref(self)

        @pipeline_thread.invoke_on_pipeline_thread_nowait
        def on_reconnect_timer_expired():
            this = self_weakref()
            logger.debug(
                "{}: Reconnect timer expired. State is {}.".format(
                    self.name, self.nucleus.connection_state
                )
            )
            # Clear the reconnect timer here first and foremost so it doesn't accidentally
            # get left around somehow. Don't use the _clear_reconnect_timer method, as the timer
            # has expired, and thus cannot be cancelled.
            this.reconnect_timer = None

            if this.nucleus.connection_state is ConnectionState.DISCONNECTED:
                # We are still disconnected, so reconnect

                # NOTE: Because any reconnect timer would have been cancelled upon a manual
                # disconnect, there is no way this block could be executing if we were happy
                # with our DISCONNECTED state.
                logger.debug("{}: Starting reconnection".format(this.name))
                logger.debug(
                    "{}: State changes {} -> CONNECTING. Sending new connect op down in reconnect attempt".format(
                        self.name, self.nucleus.connection_state
                    )
                )
                self.nucleus.connection_state = ConnectionState.CONNECTING
                this._reconnect()
            elif this.nucleus.connection_state in self.intermediate_states:
                # If another connection op is in progress, just wait and try again later to avoid
                # any extra confusion (i.e. punt the reconnection)
                logger.debug(
                    "{}: Other connection operation in-progress, setting a new reconnection timer".format(
                        this.name
                    )
                )
                this._start_reconnect_timer(
                    this.nucleus.pipeline_configuration.connection_retry_interval
                )
            else:
                logger.debug(
                    "{}: Unexpected state reached ({}) after reconnection timer expired".format(
                        this.name, this.nucleus.connection_state
                    )
                )

        self.reconnect_timer = threading.Timer(delay, on_reconnect_timer_expired)
        self.reconnect_timer.start()

    @pipeline_thread.runs_on_pipeline_thread
    def _clear_reconnect_timer(self):
        """
        Clear any previous reconnect timer
        """
        if self.reconnect_timer:
            logger.debug("{}: clearing reconnect timer".format(self.name))
            self.reconnect_timer.cancel()
            self.reconnect_timer = None



"""Azure IoT Hub Device SDK Pipeline

This package provides a protocol pipeline for use with the Azure IoT Hub Device SDK.

INTERNAL USAGE ONLY
"""

from .mqtt_pipeline import MQTTPipeline  # noqa: F401
from .http_pipeline import HTTPPipeline  # noqa: F401
from .config import IoTHubPipelineConfig  # noqa: F401



"""Azure IoT Hub Device SDK Models

This package provides object models for use within the Azure IoT Hub Device SDK.
"""

from .message import Message  # noqa: F401
from .methods import MethodRequest, MethodResponse  # noqa: F401



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains a class representing messages that are sent or received.
"""
from azure.iot.device import constant
import sys


class Message(object):
    """Represents a message to or from IoTHub

    :ivar data: The data that constitutes the payload
    :ivar custom_properties: Dictionary of custom message properties. The keys and values of these properties will always be string.
    :ivar message id: A user-settable identifier for the message used for request-reply patterns. Format: A case-sensitive string (up to 128 characters long) of ASCII 7-bit alphanumeric characters + {'-', ':', '.', '+', '%', '_', '#', '*', '?', '!', '(', ')', ',', '=', '@', ';', '$', '''}
    :ivar expiry_time_utc: Date and time of message expiration in UTC format
    :ivar correlation_id: A property in a response message that typically contains the message_id of the request, in request-reply patterns
    :ivar user_id: An ID to specify the origin of messages
    :ivar content_encoding: Content encoding of the message data. Can be 'utf-8', 'utf-16' or 'utf-32'
    :ivar content_type: Content type property used to route messages with the message-body. Can be 'application/json'
    :ivar output_name: Name of the output that the message is being sent to.
    :ivar input_name: Name of the input that the message was received on.
    """

    def __init__(
        self, data, message_id=None, content_encoding=None, content_type=None, output_name=None
    ):
        """
        Initializer for Message

        :param data: The  data that constitutes the payload
        :param str message_id: A user-settable identifier for the message used for request-reply patterns. Format: A case-sensitive string (up to 128 characters long) of ASCII 7-bit alphanumeric characters + {'-', ':', '.', '+', '%', '_', '#', '*', '?', '!', '(', ')', ',', '=', '@', ';', '$', '''}
        :param str content_encoding: Content encoding of the message data. Other values can be utf-16' or 'utf-32'
        :param str content_type: Content type property used to routes with the message body.
        :param str output_name: Name of the output that the is being sent to.
        """
        self.data = data
        self.custom_properties = {}
        self.message_id = message_id
        self.expiry_time_utc = None
        self.correlation_id = None
        self.user_id = None
        self.content_encoding = content_encoding
        self.content_type = content_type
        self.output_name = output_name
        self.input_name = None
        self.ack = None
        self._iothub_interface_id = None

    @property
    def iothub_interface_id(self) -> str:
        return self._iothub_interface_id

    def set_as_security_message(self):
        """
        Set the message as a security message.

        This is a provisional API. Functionality not yet guaranteed.
        """
        self._iothub_interface_id = constant.SECURITY_MESSAGE_INTERFACE_ID

    def __str__(self):
        return str(self.data)

    def get_size(self) -> int:
        total = 0
        total = total + sum(
            sys.getsizeof(v)
            for v in self.__dict__.values()
            if v is not None and v is not self.custom_properties
        )
        if self.custom_properties:
            total = total + sum(
                sys.getsizeof(v) for v in self.custom_properties.values() if v is not None
            )
        return total



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from typing import Any, Union, Dict, List, Tuple, Callable, Awaitable, TypeVar
from typing_extensions import TypedDict, ParamSpec


_P = ParamSpec("_P")
_R = TypeVar("_R")
FunctionOrCoroutine = Union[Callable[_P, _R], Callable[_P, Awaitable[_R]]]


# typing does not support recursion, so we must use forward references here (PEP484)
JSONSerializable = Union[
    Dict[str, "JSONSerializable"],
    List["JSONSerializable"],
    Tuple["JSONSerializable", ...],
    str,
    int,
    float,
    bool,
    None,
]
# TODO: verify that the JSON specification requires str as keys in dict. Not sure why that's defined here.


Twin = Dict[str, Dict[str, JSONSerializable]]
TwinPatch = Dict[str, JSONSerializable]


class StorageInfo(TypedDict):
    correlationId: str
    hostName: str
    containerName: str
    blobName: str
    sasToken: str


ProvisioningPayload = Union[Dict[str, Any], str, int]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
from azure.iot.device.common.evented_callback import EventedCallback
from azure.iot.device.common.pipeline import (
    pipeline_nucleus,
    pipeline_stages_base,
    pipeline_ops_base,
    pipeline_stages_mqtt,
    pipeline_exceptions,
)
from . import (
    constant,
    pipeline_stages_iothub,
    pipeline_events_iothub,
    pipeline_ops_iothub,
    pipeline_stages_iothub_mqtt,
)

logger = logging.getLogger(__name__)


class MQTTPipeline(object):
    def __init__(self, pipeline_configuration):
        """
        Constructor for instantiating a pipeline adapter object
        :param auth_provider: The authentication provider
        :param pipeline_configuration: The configuration generated based on user inputs
        """

        self.feature_enabled = {
            constant.C2D_MSG: False,
            constant.INPUT_MSG: False,
            constant.METHODS: False,
            constant.TWIN: False,
            constant.TWIN_PATCHES: False,
        }

        # Handlers - Will be set by Client after instantiation of this object
        self.on_connected = None
        self.on_disconnected = None
        self.on_new_sastoken_required = None
        self.on_background_exception = None

        self.on_c2d_message_received = None
        self.on_input_message_received = None
        self.on_method_request_received = None
        self.on_twin_patch_received = None

        # Contains data and information shared globally within the pipeline
        self._nucleus = pipeline_nucleus.PipelineNucleus(pipeline_configuration)

        self._pipeline = (
            #
            # The root is always the root.  By definition, it's the first stage in the pipeline.
            #
            pipeline_stages_base.PipelineRootStage(self._nucleus)
            #
            # SasTokenStage comes near the root by default because it should be as close
            # to the top of the pipeline as possible, and does not need to be after anything.
            #
            .append_stage(pipeline_stages_base.SasTokenStage())
            #
            # EnsureDesiredPropertiesStage needs to be above TwinRequestResponseStage because it
            # sends GetTwinOperation ops and that stage handles those ops.
            #
            .append_stage(pipeline_stages_iothub.EnsureDesiredPropertiesStage())
            #
            # TwinRequestResponseStage comes near the root by default because it doesn't need to be
            # after anything
            #
            .append_stage(pipeline_stages_iothub.TwinRequestResponseStage())
            #
            # CoordinateRequestAndResponseStage needs to be after TwinRequestResponseStage because
            # TwinRequestResponseStage creates the request ops that CoordinateRequestAndResponseStage
            # is coordinating.  It needs to be before IoTHubMQTTTranslationStage because that stage
            # operates on ops that CoordinateRequestAndResponseStage produces
            #
            .append_stage(pipeline_stages_base.CoordinateRequestAndResponseStage())
            #
            # IoTHubMQTTTranslationStage comes here because this is the point where we can translate
            # all operations directly into MQTT.  After this stage, only pipeline_stages_base stages
            # are allowed because IoTHubMQTTTranslationStage removes all the IoTHub-ness from the ops
            #
            .append_stage(pipeline_stages_iothub_mqtt.IoTHubMQTTTranslationStage())
            #
            # AutoConnectStage comes here because only MQTT ops have the need_connection flag set
            # and this is the first place in the pipeline where we can guarantee that all network
            # ops are MQTT ops.
            #
            .append_stage(pipeline_stages_base.AutoConnectStage())
            #
            # ConnectionStateStage needs to be after AutoConnectStage because the AutoConnectStage
            # can create ConnectOperations and we (may) want to queue connection related operations
            # in the ConnectionStateStage
            #
            .append_stage(pipeline_stages_base.ConnectionStateStage())
            #
            # RetryStage needs to be near the end because it's retrying low-level MQTT operations.
            #
            .append_stage(pipeline_stages_base.RetryStage())
            #
            # OpTimeoutStage needs to be after RetryStage because OpTimeoutStage returns the timeout
            # errors that RetryStage is watching for.
            #
            .append_stage(pipeline_stages_base.OpTimeoutStage())
            #
            # MQTTTransportStage needs to be at the very end of the pipeline because this is where
            # operations turn into network traffic
            #
            .append_stage(pipeline_stages_mqtt.MQTTTransportStage())
        )

        # Define behavior for domain-specific events
        def _on_pipeline_event(event):
            if isinstance(event, pipeline_events_iothub.C2DMessageEvent):
                if self.on_c2d_message_received:
                    self.on_c2d_message_received(event.message)
                else:
                    logger.debug("C2D message event received with no handler.  dropping.")

            elif isinstance(event, pipeline_events_iothub.InputMessageEvent):
                if self.on_input_message_received:
                    self.on_input_message_received(event.message)
                else:
                    logger.debug("input message event received with no handler.  dropping.")

            elif isinstance(event, pipeline_events_iothub.MethodRequestEvent):
                if self.on_method_request_received:
                    self.on_method_request_received(event.method_request)
                else:
                    logger.debug("Method request event received with no handler. Dropping.")

            elif isinstance(event, pipeline_events_iothub.TwinDesiredPropertiesPatchEvent):
                if self.on_twin_patch_received:
                    self.on_twin_patch_received(event.patch)
                else:
                    logger.debug("Twin patch event received with no handler. Dropping.")

            else:
                logger.debug("Dropping unknown pipeline event {}".format(event.name))

        def _on_connected():
            if self.on_connected:
                self.on_connected()
            else:
                logger.debug("IoTHub Pipeline was connected, but no handler was set")

        def _on_disconnected():
            if self.on_disconnected:
                self.on_disconnected()
            else:
                logger.debug("IoTHub Pipeline was disconnected, but no handler was set")

        def _on_new_sastoken_required():
            if self.on_new_sastoken_required:
                self.on_new_sastoken_required()
            else:
                logger.debug("IoTHub Pipeline requires new SASToken, but no handler was set")

        def _on_background_exception(e):
            if self.on_background_exception:
                self.on_background_exception(e)
            else:
                logger.debug(
                    "IoTHub Pipeline experienced background exception, but no handler was set"
                )

        # Set internal event handlers
        self._pipeline.on_pipeline_event_handler = _on_pipeline_event
        self._pipeline.on_connected_handler = _on_connected
        self._pipeline.on_disconnected_handler = _on_disconnected
        self._pipeline.on_new_sastoken_required_handler = _on_new_sastoken_required
        self._pipeline.on_background_exception_handler = _on_background_exception

        # Initialize the pipeline
        callback = EventedCallback()
        op = pipeline_ops_base.InitializePipelineOperation(callback=callback)
        self._pipeline.run_op(op)
        callback.wait_for_completion()

        # Set the running flag
        self._running = True

    def _verify_running(self):
        if not self._running:
            raise pipeline_exceptions.PipelineNotRunning(
                "Cannot execute method - Pipeline is not running"
            )

    def shutdown(self, callback):
        """Shut down the pipeline and clean up any resources.

        Once shut down, making any further calls on the pipeline will result in a
        PipelineNotRunning exception being raised.

        There is currently no way to resume pipeline functionality once shutdown has occurred.

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.PipelineNotRunning` if the
            pipeline has already been shut down

        The shutdown process itself is not expected to fail under any normal condition, but if it
        does, exceptions are not "raised", but rather, returned via the "error" parameter when
        invoking "callback".
        """
        self._verify_running()
        logger.debug("Commencing shutdown of pipeline")

        def on_complete(op, error):
            if not error:
                # Only set the pipeline to not be running if the op was successful
                self._running = False
            callback(error=error)

        # NOTE: While we do run this operation, its functionality is incomplete. Some stages still
        # need a response to this operation implemented. Additionally, there are other pipeline
        # constructs other than Stages (e.g. Operations) which may have timers attached. These are
        # lesser issues, but should be addressed at some point.
        # TODO: Truly complete the shutdown implementation
        self._pipeline.run_op(pipeline_ops_base.ShutdownPipelineOperation(callback=on_complete))

    def connect(self, callback):
        """
        Connect to the service.

        :param callback: callback which is called when the connection attempt is complete.

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.PipelineNotRunning` if the
            pipeline has previously been shut down

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.OperationTimeout`
        """
        self._verify_running()
        logger.debug("Starting ConnectOperation on the pipeline")

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(pipeline_ops_base.ConnectOperation(callback=on_complete))

    def disconnect(self, callback):
        """
        Disconnect from the service.

        Note that even if this fails for some reason, the client will be in a disconnected state.

        :param callback: callback which is called when the disconnection is complete.

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.PipelineNotRunning` if the
            pipeline has previously been shut down

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        """
        self._verify_running()
        logger.debug("Starting DisconnectOperation on the pipeline")

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(pipeline_ops_base.DisconnectOperation(callback=on_complete))

    def reauthorize_connection(self, callback):
        """
        Reauthorize connection to the service by disconnecting and then reconnecting using
        fresh credentials.

        This can be called regardless of connection state. If successful, the client will be
        connected. If unsuccessful, the client will be disconnected.

        :param callback: callback which is called when the reauthorization attempt is complete.

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.PipelineNotRunning` if the
            pipeline has previously been shut down

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.OperationTimeout`
        """
        self._verify_running()
        logger.debug("Starting ReauthorizeConnectionOperation on the pipeline")

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_base.ReauthorizeConnectionOperation(callback=on_complete)
        )

    def send_message(self, message, callback):
        """
        Send a telemetry message to the service.

        :param message: message to send.
        :param callback: callback which is called when the message publish attempt is complete.

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.PipelineNotRunning` if the
            pipeline has previously been shut down

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.NoConnectionError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`

        The following exceptions can be returned via the "error" parameter only if auto-connect
        is enabled in the pipeline configuration:

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.OperationTimeout`
        """
        self._verify_running()
        logger.debug("Starting SendD2CMessageOperation on the pipeline")

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_iothub.SendD2CMessageOperation(message=message, callback=on_complete)
        )

    def send_output_message(self, message, callback):
        """
        Send an output message to the service.

        :param message: message to send.
        :param callback: callback which is called when the message publish attempt is complete.

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.PipelineNotRunning` if the
            pipeline has previously been shut down

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.NoConnectionError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`

        The following exceptions can be returned via the "error" parameter only if auto-connect
        is enabled in the pipeline configuration:

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.OperationTimeout`
        """
        self._verify_running()
        logger.debug("Starting SendOutputMessageOperation on the pipeline")

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_iothub.SendOutputMessageOperation(message=message, callback=on_complete)
        )

    def send_method_response(self, method_response, callback):
        """
        Send a method response to the service.

        :param method_response: the method response to send
        :param callback: callback which is called when response publish attempt is complete.

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.PipelineNotRunning` if the
            pipeline has previously been shut down

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.NoConnectionError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`

        The following exceptions can be returned via the "error" parameter only if auto-connect
        is enabled in the pipeline configuration:

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.OperationTimeout`
        """
        self._verify_running()
        logger.debug("Starting SendMethodResponseOperation on the pipeline")

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_iothub.SendMethodResponseOperation(
                method_response=method_response, callback=on_complete
            )
        )

    def get_twin(self, callback):
        """
        Send a request for a full twin to the service.

        :param callback: callback which is called when request attempt is complete.
            This callback should have two parameters.  On success, this callback is called with the
            requested twin and error=None.  On failure, this callback is called with None for the
            requested win and error set to the cause of the failure.

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.PipelineNotRunning` if the
            pipeline has previously been shut down

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.NoConnectionError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`

        The following exceptions can be returned via the "error" parameter only if auto-connect
        is enabled in the pipeline configuration:

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.OperationTimeout`
        """
        self._verify_running()
        logger.debug("Starting GetTwinOperation on the pipeline")

        def on_complete(op, error):
            if error:
                callback(error=error, twin=None)
            else:
                callback(twin=op.twin)

        self._pipeline.run_op(pipeline_ops_iothub.GetTwinOperation(callback=on_complete))

    def patch_twin_reported_properties(self, patch, callback):
        """
        Send a patch for a twin's reported properties to the service.

        :param patch: the reported properties patch to send
        :param callback: callback which is called when the request attempt is complete.

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.PipelineNotRunning` if the
            pipeline has previously been shut down

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.NoConnectionError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`

        The following exceptions can be returned via the "error" parameter only if auto-connect
        is enabled in the pipeline configuration:

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.OperationTimeout`
        """
        self._verify_running()
        logger.debug("Starting PatchTwinReportedPropertiesOperation on the pipeline")

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_iothub.PatchTwinReportedPropertiesOperation(
                patch=patch, callback=on_complete
            )
        )

    # NOTE: Currently, this operation will retry itself indefinitely in the case of timeout
    def enable_feature(self, feature_name, callback):
        """
        Enable the given feature by subscribing to the appropriate topics.

        :param feature_name: one of the feature name constants from constant.py
        :param callback: callback which is called when the feature is enabled

        :raises: ValueError if feature_name is invalid
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.PipelineNotRunning` if the
            pipeline has previously been shut down

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.NoConnectionError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`

        The following exceptions can be returned via the "error" parameter only if auto-connect
        is enabled in the pipeline configuration:

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        """
        self._verify_running()
        logger.debug("enable_feature {} called".format(feature_name))
        if feature_name not in self.feature_enabled:
            raise ValueError("Invalid feature_name")
        # TODO: What about if the feature is already enabled?

        def on_complete(op, error):
            if error:
                logger.warning(
                    "Subscribe for {} failed.  Not enabling feature".format(feature_name)
                )
            else:
                self.feature_enabled[feature_name] = True
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_base.EnableFeatureOperation(
                feature_name=feature_name, callback=on_complete
            )
        )

    # NOTE: Currently, this operation will retry itself indefinitely in the case of timeout
    def disable_feature(self, feature_name, callback):
        """
        Disable the given feature by subscribing to the appropriate topics.
        :param callback: callback which is called when the feature is disabled

        :param feature_name: one of the feature name constants from constant.py

        :raises: ValueError if feature_name is invalid
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.PipelineNotRunning` if the
            pipeline has previously been shut down

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.NoConnectionError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`

        The following exceptions can be returned via the "error" parameter only if auto-connect
        is enabled in the pipeline configuration:

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        """
        self._verify_running()
        logger.debug("disable_feature {} called".format(feature_name))
        if feature_name not in self.feature_enabled:
            raise ValueError("Invalid feature_name")
        # TODO: What about if the feature is already disabled?

        def on_complete(op, error):
            if error:
                logger.warning(
                    "Error occurred while disabling feature. Unclear if subscription for {} is still alive or not".format(
                        feature_name
                    )
                )

            # No matter what, mark the feature as disabled, even if there was an error.
            # This is safer than only marking it disabled upon operation success, because an op
            # could fail after successfully doing the network operations to change the subscription
            # state, and then we would be stuck in a bad state.
            self.feature_enabled[feature_name] = False
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_base.DisableFeatureOperation(
                feature_name=feature_name, callback=on_complete
            )
        )

    @property
    def pipeline_configuration(self):
        """
        Pipeline Configuration for the pipeline. Note that while a new config object cannot be
        provided (read-only), the values stored in the config object CAN be changed.
        """
        return self._nucleus.pipeline_configuration

    @property
    def connected(self):
        """
        Read-only property to indicate if the transport is connected or not.
        """
        return self._nucleus.connected



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
This module represents proxy options to enable sending traffic through proxy servers.
"""
import socks

string_to_socks_constant_map = {"HTTP": socks.HTTP, "SOCKS4": socks.SOCKS4, "SOCKS5": socks.SOCKS5}

socks_constant_to_string_map = {socks.HTTP: "HTTP", socks.SOCKS4: "SOCKS4", socks.SOCKS5: "SOCKS5"}


class ProxyOptions(object):
    """
    A class containing various options to send traffic through proxy servers by enabling
    proxying of MQTT connection.
    """

    def __init__(
        self, proxy_type, proxy_addr, proxy_port, proxy_username=None, proxy_password=None
    ):
        """
        Initializer for proxy options.
        :param str proxy_type: The type of the proxy server. This can be one of three possible choices: "HTTP", "SOCKS4", or "SOCKS5"
        :param str proxy_addr: IP address or DNS name of proxy server
        :param int proxy_port: The port of the proxy server. Defaults to 1080 for socks and 8080 for http.
        :param str proxy_username: (optional) username for SOCKS5 proxy, or userid for SOCKS4 proxy.This parameter is ignored if an HTTP server is being used.
         If it is not provided, authentication will not be used (servers may accept unauthenticated requests).
        :param str proxy_password: (optional) This parameter is valid only for SOCKS5 servers and specifies the respective password for the username provided.
        """
        (self._proxy_type, self._proxy_type_socks) = format_proxy_type(proxy_type)
        self._proxy_addr = proxy_addr
        self._proxy_port = int(proxy_port)
        self._proxy_username = proxy_username
        self._proxy_password = proxy_password

    @property
    def proxy_type(self):
        return self._proxy_type

    @property
    def proxy_type_socks(self):
        return self._proxy_type_socks

    @property
    def proxy_address(self):
        return self._proxy_addr

    @property
    def proxy_port(self):
        return self._proxy_port

    @property
    def proxy_username(self):
        return self._proxy_username

    @property
    def proxy_password(self):
        return self._proxy_password


def format_proxy_type(proxy_type):
    """Returns a tuple of formats for proxy type (string, socks library constant)"""
    try:
        return (proxy_type, string_to_socks_constant_map[proxy_type])
    except KeyError:
        # Backwards compatibility for when we used the socks library constants in the API
        try:
            return (socks_constant_to_string_map[proxy_type], proxy_type)
        except KeyError:
            raise ValueError("Invalid Proxy Type")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
""" This module contains functions of managing event loops for the IoTHub client
"""
import asyncio
import threading
import logging

logger = logging.getLogger(__name__)

loops = {
    "CLIENT_HANDLER_LOOP": None,
    "CLIENT_INTERNAL_LOOP": None,
    "CLIENT_HANDLER_RUNNER_LOOP": None,
}


def _cleanup():
    """Clear all running loops and end respective threads.
    ONLY FOR TESTING USAGE
    By using this function, you can wipe all global loops.
    DO NOT USE THIS IN PRODUCTION CODE
    """
    for loop_name, loop in loops.items():
        if loop is not None:
            logger.debug("Stopping event loop - {}".format(loop_name))
            loop.call_soon_threadsafe(loop.stop)
            # NOTE: Stopping the loop will also end the thread, because the only thing keeping
            # the thread alive was the loop running
            loops[loop_name] = None


def _make_new_loop(loop_name):
    logger.debug("Creating new event loop - {}".format(loop_name))
    # Create the loop on a new Thread
    new_loop = asyncio.new_event_loop()
    loop_thread = threading.Thread(target=new_loop.run_forever)
    # Make the Thread a daemon so it will not block program exit
    loop_thread.daemon = True
    loop_thread.start()
    # Store the loop
    loops[loop_name] = new_loop


def get_client_internal_loop():
    """Return the loop for internal client operations"""
    if loops["CLIENT_INTERNAL_LOOP"] is None:
        _make_new_loop("CLIENT_INTERNAL_LOOP")
    return loops["CLIENT_INTERNAL_LOOP"]


def get_client_handler_runner_loop():
    """Return the loop for handler runners"""
    if loops["CLIENT_HANDLER_RUNNER_LOOP"] is None:
        _make_new_loop("CLIENT_HANDLER_RUNNER_LOOP")
    return loops["CLIENT_HANDLER_RUNNER_LOOP"]


def get_client_handler_loop():
    """Return the loop for invoking user-provided handlers on the client"""
    # TODO: Try and store the user loop somehow
    if loops["CLIENT_HANDLER_LOOP"] is None:
        _make_new_loop("CLIENT_HANDLER_LOOP")
    return loops["CLIENT_HANDLER_LOOP"]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import threading
import logging
import traceback

logger = logging.getLogger(__name__)


class EventedCallback(object):
    """
    A sync callback whose completion can be waited upon.
    """

    def __init__(self, return_arg_name=None):
        """
        Creates an instance of an EventedCallback.

        """
        # LBYL because this mistake doesn't cause an exception until the callback
        # which is much later and very difficult to trace back to here.
        if return_arg_name and not isinstance(return_arg_name, str):
            raise TypeError("internal error: return_arg_name must be a string")

        self.completion_event = threading.Event()
        self.exception = None
        self.result = None

        def wrapping_callback(*args, **kwargs):
            if "error" in kwargs and kwargs["error"]:
                self.exception = kwargs["error"]
            elif return_arg_name:
                if return_arg_name in kwargs:
                    self.result = kwargs[return_arg_name]
                else:
                    raise TypeError(
                        "internal error: expected argument with name '{}', did not get".format(
                            return_arg_name
                        )
                    )

            if self.exception:
                # Do not use exc_info parameter on logger.* calls.  This causes pytest to save the traceback which saves stack frames which shows up as a leak
                logger.info("Callback completed with error {}".format(self.exception))
                logger.info(traceback.format_exc())
            else:
                logger.debug("Callback completed with result {}".format(self.result))

            self.completion_event.set()

        self.callback = wrapping_callback

    def __call__(self, *args, **kwargs):
        """
        Calls the callback.
        """
        self.callback(*args, **kwargs)

    def wait_for_completion(self, *args, **kwargs):
        """
        Wait for the callback to be called, and return the results.
        """
        self.completion_event.wait(*args, **kwargs)

        if self.exception:
            raise self.exception
        else:
            return self.result



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from threading import Thread, Event
import time

# NOTE: The Alarm class is very similar, but fundamentally different from threading.Timer.
# Beyond just the input format difference (a specific time for Alarm vs. an interval for Timer),
# the manner in which they keep time is different. A Timer will only tick towards its interval
# while the system is running, but an Alarm will go off at a specific time, so system sleep will
# not throw off the timekeeping. In the case that the Alarm time occurs while the system is asleep
# the Alarm will trigger upon system wake.


class Alarm(Thread):
    """Call a function at a specified time"""

    def __init__(self, alarm_time, function, args=None, kwargs=None):
        Thread.__init__(self)
        self.alarm_time = alarm_time
        self.function = function
        self.args = args if args is not None else []
        self.kwargs = kwargs if kwargs is not None else {}
        self.finished = Event()

    def cancel(self):
        """Stop the alarm if it hasn't finished yet."""
        self.finished.set()

    def run(self):
        """Method representing the thread's activity.
        Overrides the method inherited from Thread.
        Will invoke the Alarm's given function at the given alarm time (accurate within 1 second)
        """
        while not self.finished.is_set() and time.time() < self.alarm_time:
            self.finished.wait(1)

        if not self.finished.is_set():
            self.function(*self.args, **self.kwargs)
        self.finished.set()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
""" This module contains the manager for handler methods used by the aio clients"""

import asyncio
import logging
import inspect
import concurrent.futures
from azure.iot.device.common import handle_exceptions
from azure.iot.device.iothub.sync_handler_manager import (
    AbstractHandlerManager,
    HandlerManagerException,
    HandlerRunnerKillerSentinel,
    CLIENT_EVENT,
)
from . import loop_management

logger = logging.getLogger(__name__)


class AsyncHandlerManager(AbstractHandlerManager):
    """Handler manager for use with asynchronous clients"""

    async def _receiver_handler_runner(self, inbox, handler_name):
        """Run infinite loop that waits for an inbox to receive an object from it, then calls
        the handler with that object
        """
        logger.debug("HANDLER RUNNER ({}): Starting runner".format(handler_name))

        # Define a callback that can handle errors in the ThreadPoolExecutor
        _handler_callback = self._generate_callback_for_handler("CLIENT_EVENT")

        # ThreadPool used for running handler functions. By invoking handlers in a separate thread
        # we can be safe knowing that customer code that has performance issues does not block
        # client code. Note that the ThreadPool is only used for handler FUNCTIONS (coroutines are
        # invoked on a dedicated event loop + thread)
        tpe = concurrent.futures.ThreadPoolExecutor(max_workers=4)
        while True:
            handler_arg = await inbox.get()
            if isinstance(handler_arg, HandlerRunnerKillerSentinel):
                # Exit the runner when a HandlerRunnerKillerSentinel is found
                logger.debug(
                    "HANDLER RUNNER ({}): HandlerRunnerKillerSentinel found in inbox. Exiting.".format(
                        handler_name
                    )
                )
                tpe.shutdown()
                break
            # NOTE: we MUST use getattr here using the handler name, as opposed to directly passing
            # the handler in order for the handler to be able to be updated without cancelling
            # the running task created for this coroutine
            handler = getattr(self, handler_name)
            logger.debug("HANDLER RUNNER ({}): Invoking handler".format(handler_name))
            if inspect.iscoroutinefunction(handler):
                # Run coroutine on a dedicated event loop for handler invocations
                # TODO: Can we call this on the user loop instead?
                handler_loop = loop_management.get_client_handler_loop()
                fut = asyncio.run_coroutine_threadsafe(handler(handler_arg), handler_loop)
                # Free up this object so the garbage collector can free it if necessary. If we don't
                # do this, we end up keeping this object alive until the next event arrives, which
                # might be a long time. Tests would flag this as a memory leak if that happened.
                del handler_arg
                fut.add_done_callback(_handler_callback)
            else:
                # Run function directly in ThreadPool
                fut = tpe.submit(handler, handler_arg)
                # Free up this object so the garbage collector can free it if necessary. If we don't
                # do this, we end up keeping this object alive until the next event arrives, which
                # might be a long time. Tests would flag this as a memory leak if that happened.
                del handler_arg
                fut.add_done_callback(_handler_callback)

    async def _client_event_handler_runner(self):
        """Run infinite loop that waits for the client event inbox to receive an event from it,
        then calls the handler that corresponds to that event
        """
        logger.debug("HANDLER RUNNER (CLIENT EVENT): Starting runner")
        _handler_callback = self._generate_callback_for_handler("CLIENT_EVENT")

        # ThreadPool used for running handler functions. By invoking handlers in a separate thread
        # we can be safe knowing that customer code that has performance issues does not block
        # client code. Note that the ThreadPool is only used for handler FUNCTIONS (coroutines are
        # invoked on a dedicated event loop + thread)
        tpe = concurrent.futures.ThreadPoolExecutor(max_workers=4)
        event_inbox = self._inbox_manager.get_client_event_inbox()
        while True:
            event = await event_inbox.get()
            if isinstance(event, HandlerRunnerKillerSentinel):
                # Exit the runner when a HandlerRunnerKillerSentinel is found
                logger.debug(
                    "HANDLER RUNNER (CLIENT EVENT): HandlerRunnerKillerSentinel found in event queue. Exiting."
                )
                tpe.shutdown()
                break
            handler = self._get_handler_for_client_event(event.name)
            if handler is not None:
                logger.debug(
                    "HANDLER RUNNER (CLIENT EVENT): {} event received. Invoking {} handler".format(
                        event, handler
                    )
                )
                if inspect.iscoroutinefunction(handler):
                    # Run a coroutine on a dedicated event loop for handler invocations
                    # TODO: Can we call this on the user loop instead?
                    handler_loop = loop_management.get_client_handler_loop()
                    fut = asyncio.run_coroutine_threadsafe(
                        handler(*event.args_for_user), handler_loop
                    )
                    # Free up this object so the garbage collector can free it if necessary. If we don't
                    # do this, we end up keeping this object alive until the next event arrives, which
                    # might be a long time. Tests would flag this as a memory leak if that happened.
                    del event
                    fut.add_done_callback(_handler_callback)
                else:
                    # Run a function directly in ThreadPool
                    fut = tpe.submit(handler, *event.args_for_user)
                    # Free up this object so the garbage collector can free it if necessary. If we don't
                    # do this, we end up keeping this object alive until the next event arrives, which
                    # might be a long time. Tests would flag this as a memory leak if that happened.
                    del event
                    fut.add_done_callback(_handler_callback)
            else:
                logger.debug(
                    "No handler for event {} set. Skipping handler invocation".format(event)
                )

    def _start_handler_runner(self, handler_name):
        """Create, and store a task for running a handler"""
        # Run the handler runner on a dedicated event loop for handler runners so as to be
        # isolated from all other client activities
        runner_loop = loop_management.get_client_handler_runner_loop()

        # Client Event handler flow
        if handler_name == CLIENT_EVENT:
            if self._client_event_runner is not None:
                # This branch of code should NOT be reachable due to checks prior to the invocation
                # of this method. The branch exists for safety.
                raise HandlerManagerException(
                    "Cannot create thread for handler runner: {}. Runner thread already exists".format(
                        handler_name
                    )
                )
            # Client events share a handler
            coro = self._client_event_handler_runner()
            future = asyncio.run_coroutine_threadsafe(coro, runner_loop)
            # Store the future
            self._client_event_runner = future

        # Receiver handler flow
        else:
            if self._receiver_handler_runners[handler_name] is not None:
                # This branch of code should NOT be reachable due to checks prior to the invocation
                # of this method. The branch exists for safety.
                raise HandlerManagerException(
                    "Cannot create task for handler runner: {}. Task already exists".format(
                        handler_name
                    )
                )
            # Each receiver handler gets its own runner
            inbox = self._get_inbox_for_receive_handler(handler_name)
            coro = self._receiver_handler_runner(inbox, handler_name)
            future = asyncio.run_coroutine_threadsafe(coro, runner_loop)
            # Store the future
            self._receiver_handler_runners[handler_name] = future

        _handler_runner_callback = self._generate_callback_for_handler_runner(handler_name)
        future.add_done_callback(_handler_runner_callback)

    def _stop_receiver_handler_runner(self, handler_name):
        """Stop and remove a handler runner task.
        All pending items in the corresponding inbox will be handled by the handler before stoppage.
        """
        logger.debug(
            "Adding HandlerRunnerKillerSentinel to inbox corresponding to {} handler runner".format(
                handler_name
            )
        )
        inbox = self._get_inbox_for_receive_handler(handler_name)
        inbox.put(HandlerRunnerKillerSentinel())

        # Wait for Handler Runner to end due to the sentinel
        logger.debug("Waiting for {} handler runner to exit...".format(handler_name))
        future = self._receiver_handler_runners[handler_name]
        future.result()
        # Stop tracking the task since it is now complete
        self._receiver_handler_runners[handler_name] = None
        logger.debug("Handler runner for {} has been stopped".format(handler_name))

    def _stop_client_event_handler_runner(self):
        """Stop and remove a handler task.
        All pending items in the client event queue will be handled by handlers (if they exist)
        before stoppage.
        """
        logger.debug("Adding HandlerRunnerKillerSentinel to client event queue")
        event_inbox = self._inbox_manager.get_client_event_inbox()
        event_inbox.put(HandlerRunnerKillerSentinel())

        # Wait for Handler Runner to end due to the stop command
        logger.debug("Waiting for client event handler runner to exit...")
        future = self._client_event_runner
        future.result()
        # Stop tracking the task since it is now complete
        self._client_event_runner = None
        logger.debug("Handler runner for client events has been stopped")

    def _generate_callback_for_handler_runner(self, handler_name):
        """Define a callback that can handle errors during handler runner execution"""

        def handler_runner_callback(completed_future):
            try:
                e = completed_future.exception(timeout=0)
            except Exception as raised_e:
                # This shouldn't happen because cancellation or timeout shouldn't occur...
                # But just in case...
                new_err = HandlerManagerException(
                    "HANDLER RUNNER ({}): Unable to retrieve exception data from incomplete task".format(
                        handler_name
                    )
                )
                new_err.__cause__ = raised_e
                handle_exceptions.handle_background_exception(new_err)
            else:
                if e:
                    # If this branch is reached something has gone SERIOUSLY wrong.
                    # We must log the error, and then restart the runner so that the program
                    # does not enter an invalid state
                    new_err = HandlerManagerException(
                        "HANDLER RUNNER ({}): Unexpected error during task".format(handler_name),
                    )
                    new_err.__cause__ = e
                    handle_exceptions.handle_background_exception(new_err)
                    # Clear the tracked runner, and start a new one
                    logger.debug("HANDLER RUNNER ({}): Restarting handler runner")
                    self._receiver_handler_runners[handler_name] = None
                    self._start_handler_runner(handler_name)
                else:
                    logger.debug(
                        "HANDLER RUNNER ({}): Task successfully completed without exception".format(
                            handler_name
                        )
                    )

        return handler_runner_callback



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module provides patches used to dynamically modify items from the libraries"""

import inspect
import logging
from typing import Dict

logger = logging.getLogger(__name__)

# This dict will be used as a scope for imports and defs in add_shims_for_inherited_methods
# in order to keep them out of the global scope of this module.
shim_scope : Dict[str, str] = {}


def add_shims_for_inherited_methods(target_class):
    """Dynamically add overriding, pass-through shim methods for all public inherited methods
    on a child class, which simply call into the parent class implementation of the same method.

    These shim methods will include the same docstrings as the method from the parent class.

    This currently only works for Python 3.5+

    Using DEBUG logging will allow you to see output of all dynamic operations that occur within
    for debugging purposes.

    :param target_class: The child class to add shim methods to
    """

    # Depending on how the method was defined, it could be either a function or a method.
    # Thus we need to find the union of the two sets.
    # Here instance methods are considered functions because they are not yet bound to an instance
    # of the class. Classmethods on the other hand, are already bound, and show up as methods.
    # It also is worth noting that async functions/methods ARE picked up by this introspection.
    class_functions = inspect.getmembers(target_class, predicate=inspect.isfunction)
    class_methods = inspect.getmembers(target_class, predicate=inspect.ismethod)
    all_methods = class_functions + class_methods

    # This list of attributes gives us a lot of information, but we only are using it to get
    # the defining class of a given method.
    class_attributes = inspect.classify_class_attrs(target_class)

    # We must alias class names to prevent naming collisions when this fn is called multiple times
    # with classes that share a name. If we've already used this classname, add trailing underscore(s)
    classname_alias = target_class.__name__
    while classname_alias in shim_scope:
        classname_alias += "_"

    # Import the class we're adding methods to, so that functions defined in this scope can use super()
    class_module = inspect.getmodule(target_class)
    import_cmdstr = "from {module} import {target_class} as {alias}".format(
        module=class_module.__name__, target_class=target_class.__name__, alias=classname_alias
    )
    logger.debug("exec: " + import_cmdstr)
    # exec(import_cmdstr, shim_scope)

    for method in all_methods:
        method_name = method[0]
        method_obj = method[1]
        # We can index on 0 here because the list comprehension will always be exactly 1 element
        method_attribute = [att for att in class_attributes if att.name == method_name][0]
        # The object of the class where the method was originally defined.
        originating_class_obj = method_attribute.defining_class

        # Create a shim method for all public methods inherited from a parent class
        if method_name[0] != "_" and originating_class_obj != target_class:

            method_sig = inspect.signature(method_obj)
            sig_params = method_sig.parameters

            # Bound methods (i.e. classmethods) remove the first parameter (i.e. cls)
            # so we need to add it back
            if inspect.ismethod(method_obj):
                complete_params = []
                complete_params.append(
                    inspect.Parameter("cls", inspect.Parameter.POSITIONAL_OR_KEYWORD)
                )
                complete_params += list(sig_params.values())
                method_sig = method_sig.replace(parameters=complete_params)

            # Since neither "self" nor "cls" are used in invocation, we need to create a new
            # invocation signature without them
            invoke_params_list = []
            for param in sig_params.values():
                if param.name != "self" and param.name != "cls":
                    # Set the parameter to empty (since we use this in an invocation, not a signature)
                    new_param = param.replace(default=inspect.Parameter.empty)
                    invoke_params_list.append(new_param)
            invoke_params = method_sig.replace(parameters=invoke_params_list)

            # Choose syntactical variants
            if inspect.ismethod(method_obj):
                obj_or_type = "cls"  # Use 'cls' to invoke super() for classmethods
            else:
                obj_or_type = "self"  # Use 'self' to invoke super() for instance methods
            if inspect.iscoroutine(method_obj) or inspect.iscoroutinefunction(method_obj):
                def_syntax = "async def"  # Define coroutine function/method
                ret_syntax = "return await"
            else:
                def_syntax = "def"  # Define function/method
                ret_syntax = "return"

            # Dynamically define a new shim function, with the same name, that invokes the method of the parent class
            fn_def_cmdstr = "{def_syntax} {method_name}{signature}: {ret_syntax} super({leaf_class}, {object_or_type}).{method_name}{invocation}".format(
                def_syntax=def_syntax,
                method_name=method_name,
                signature=str(method_sig),
                ret_syntax=ret_syntax,
                leaf_class=classname_alias,
                object_or_type=obj_or_type,
                invocation=str(invoke_params),
            )
            logger.debug("exec: " + fn_def_cmdstr)
            # exec(fn_def_cmdstr, shim_scope)

            # Copy the docstring from the method to the shim function
            set_doc_cmdstr = "{method_name}.__doc__ = {leaf_class}.{method_name}.__doc__".format(
                method_name=method_name, leaf_class=classname_alias
            )
            logger.debug("exec: " + set_doc_cmdstr)
            # exec(set_doc_cmdstr, shim_scope)

            # Add shim function to leaf/child class as a classmethod if the method being shimmed is a classmethod
            if inspect.ismethod(method_obj):
                attach_shim_cmdstr = (
                    "setattr({leaf_class}, '{method_name}', classmethod({method_name}))".format(
                        leaf_class=classname_alias, method_name=method_name
                    )
                )
            # Add shim function to leaf/child class as a method if the method being shimmed is an instance method
            else:
                attach_shim_cmdstr = "setattr({leaf_class}, '{method_name}', {method_name})".format(
                    leaf_class=classname_alias, method_name=method_name
                )
            logger.debug("exec: " + attach_shim_cmdstr)
            # exec(attach_shim_cmdstr, shim_scope)

    # NOTE: the __qualname__ attributes of these new shim methods are merely the method name,
    # rather than <class_name>.<method_name>, due to the scoping of the definition.
    # This shouldn't matter, but in case it does, I am documenting that fact here.

    # For properties, we have a different strategy. While with methods we dynamically created
    # redefinitions for each inherited method that called the parent class' implementation of the
    # method, here we simply set the inherited property attribute directly onto the child.
    # This will carry over all docstrings implicitly.
    # We do this because properties, while defined syntactically via methods, actually are not
    # methods directly on a class, but form a "property" object, which itself contains the
    # get and set logic. Thus our strategy for methods can't really work here.
    class_properties = inspect.getmembers(target_class, predicate=inspect.isdatadescriptor)
    for prop in class_properties:
        property_name = prop[0]
        # We can index on 0 here because the list comprehension will always be exactly 1 element
        property_attribute = [att for att in class_attributes if att.name == property_name][0]
        # The object of the class where the property was originally defined.
        originating_class_obj = property_attribute.defining_class

        # Simply redefine the same property on the leaf class if it was defined on a parent
        if property_name[0] != "_" and originating_class_obj != target_class:
            attach_property_cmdstr = (
                "setattr({leaf_class}, '{property_name}', {leaf_class}.{property_name})".format(
                    leaf_class=classname_alias, property_name=property_name
                )
            )
            logger.debug("exec: " + attach_property_cmdstr)
            # exec(attach_property_cmdstr, shim_scope)



"""Azure Device Models

This package provides object models for use within the Azure Provisioning Device SDK and Azure IoTHub Device SDK.
"""

from .x509 import X509  # noqa: F401
from .proxy_options import ProxyOptions  # noqa: F401
